--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local Service = import "Service"

local BeautyService = define ("BeautyService", Service)

LOGD("BeautyService has been defined ..., parent is", Service)

function BeautyService:_OnInitialize()
    self.service_type = Service.SERVICE_BEAUTY
    self.input_texture = nil
    self.output_texture = nil
    self.external_output_texture = nil

    self.face_beauty_1_smooth = 0.9
    self.face_beauty_1 = hl.FaceBeautyFast_New()
    self.face_beauty_1_inited = false
    hl.Class_Retain(self.face_beauty_1)

    self.face_beauty_2_whiten_factor = 0.65
    self.face_beauty_2 = hl.FaceBeautyBBQ_New()
    self.face_beauty_2_inited = false
    hl.Class_Retain(self.face_beauty_2)

    self.face_beauty_3_smooth_factor = 0.59
    self.face_beauty_3_whiten_factor = 0.35
    self.face_beauty_3 = hl.FaceBeautyGF_New()
    self.face_beauty_3_inited = false
    hl.Class_Retain(self.face_beauty_3)

    self.cur_beauty_type = hl.Hl_MSG_SCRIPT_SERVICE_BEAUTY_1
    self.cur_beauty = self.face_beauty_1

    self.frame_view_id_list = hl.U16Array()

    self.configure_dirty = false
end

function BeautyService:_Prepare()
    if self.configure_dirty ~= true then
        return
    end

    if self.cur_beauty_type == hl.Hl_MSG_SCRIPT_SERVICE_BEAUTY_1 then
        self.cur_beauty = self.face_beauty_1
        if self.face_beauty_1_inited == false then
            self.face_beauty_1:InitWithSetting(self.input_texture,
                    self.external_output_texture or self.output_texture, self.face_beauty_1_smooth)
            self.face_beauty_1_inited = true
        else
            self.face_beauty_1:SetInput(self.input_texture)
            self.face_beauty_1:SetOutput(self.external_output_texture or self.output_texture)
            self.face_beauty_1:SetSmooth(self.face_beauty_1_smooth)
        end
    elseif self.cur_beauty_type == hl.Hl_MSG_SCRIPT_SERVICE_BEAUTY_2 then
        self.cur_beauty = self.face_beauty_2
        if self.face_beauty_2_inited == false then
            self.face_beauty_2:InitWithSetting(self.input_texture,
                    self.external_output_texture or self.output_texture, self.face_beauty_2_whiten_factor)
            self.face_beauty_2_inited = true
        else
            self.face_beauty_2:SetInput(self.input_texture)
            self.face_beauty_2:SetOutput(self.external_output_texture or self.output_texture)
            self.face_beauty_2:SetBeautyWhitenFactor(self.face_beauty_2_whiten_factor)
        end
    elseif self.cur_beauty_type == hl.Hl_MSG_SCRIPT_SERVICE_BEAUTY_3 then
        self.cur_beauty = self.face_beauty_3
        if self.face_beauty_3_inited == false then
            self.face_beauty_3:InitWithSetting(self.input_texture,
                    self.external_output_texture or self.output_texture, self.face_beauty_3_smooth_factor, self.face_beauty_3_whiten_factor)
            self.face_beauty_3_inited = true
        else
            self.face_beauty_3:SetInput(self.input_texture)
            self.face_beauty_3:SetOutput(self.external_output_texture or self.output_texture)
            self.face_beauty_3:SetBeautySmoothFactor(self.face_beauty_3_smooth_factor)
            self.face_beauty_3:SetBeautyWhitenFactor(self.face_beauty_3_whiten_factor)
        end
    else
        CHECK(false, "BeautyService: _Prepare: Invalid beauty type:", self.cur_beauty_type, "!!!")
    end

    self.configure_dirty = false
end

function BeautyService:_OnStop()
    self:EnableService(Service.SERVICE_SHOW_BG, false)
end

function BeautyService:_OnStart()
    self:EnableService(Service.SERVICE_SHOW_BG, true)
end

function BeautyService:_OnUpdate(dt)
    self:_Prepare()

    if self.cur_beauty:PrepareRender() then
        local count = self.cur_beauty:GetRenderViewCount()
        self.frame_view_id_list:clear()
        for k = 1, count do
            self.frame_view_id_list:push_back(hl.GPRenderDrive_AllocFrameView())
        end
        self.cur_beauty:Render(self.frame_view_id_list)
    end
end

function BeautyService:_OnNodeEvent(msg, message_source)
    local param_num = msg:PopInt32()

    CHECK(param_num > 0, "BeautyService: _OnNodeEvent: param num:", param_num, "<= 0 !!!")

    local param_key = nil
    local param_value = nil
    for i = 1, param_num do
        param_key = msg:PopInt32()
        if param_key == hl.Hl_MSG_SCRIPT_SERVICE_BEAUTY_TYPE then
            param_value = msg:PopInt32()
            self.cur_beauty_type = param_value
            LOGD("BeautyService: _OnNodeEvent: beauty_type ->", self.cur_beauty_type)
        elseif param_key == hl.Hl_MSG_SCRIPT_SERVICE_BEAUTY_1_SMOOTH then
            param_value = msg:PopFloat()
            self.face_beauty_1_smooth = param_value
            LOGD("BeautyService: _OnNodeEvent: face_beauty_1_smooth ->", self.face_beauty_1_smooth)
        elseif param_key == hl.Hl_MSG_SCRIPT_SERVICE_BEAUTY_2_WHITEN_FACTOR then
            param_value = msg:PopFloat()
            self.face_beauty_2_whiten_factor = param_value
            LOGD("BeautyService: _OnNodeEvent: face_beauty_2_whiten_factor ->", self.face_beauty_2_whiten_factor)
        elseif param_key == hl.Hl_MSG_SCRIPT_SERVICE_BEAUTY_3_SMOOTH_FACTOR then
            param_value = msg:PopFloat()
            self.face_beauty_3_smooth_factor = param_value
            LOGD("BeautyService: _OnNodeEvent: face_beauty_3_smooth_factor ->", self.face_beauty_3_smooth_factor)
        elseif param_key == hl.Hl_MSG_SCRIPT_SERVICE_BEAUTY_3_WHITEN_FACTOR then
            param_value = msg:PopFloat()
            self.face_beauty_3_whiten_factor = param_value
            LOGD("BeautyService: _OnNodeEvent: face_beauty_3_whiten_factor ->", self.face_beauty_3_whiten_factor)
        else
            CHECK(false, "BeautyService: _OnNodeEvent: unknown param key:", param_key, "!!!")
        end
    end

    self.configure_dirty = true
end

function BeautyService:SetInputTexture(texture)
    CHECK(texture ~= nil, "BeautyService: SetInputTexture: input texture is nil !!!")
    self.input_texture = texture
    if self.output_texture == nil or
            self.output_texture:GetWidth() ~= texture:GetWidth() or
            self.output_texture:GetHeight() ~= texture:GetHeight() or
            self.output_texture:GetType() ~= texture:GetType() or
            self.output_texture:GetFormat() ~= texture:GetFormat() then

        if self.output_texture ~= nil then
            hl.Class_Release(self.output_texture)
        end

        self.output_texture = hl.RenderTexture_New()
        self.output_texture:SetWidth(self.input_texture:GetWidth())
        self.output_texture:SetHeight(self.input_texture:GetHeight())
        self.output_texture:SetType(self.input_texture:GetType())
        self.output_texture:SetFlags(hl.GPTextureFlag_kDefault2DRT)
        self.output_texture:SetFormat(self.input_texture:GetFormat())
        hl.Class_Retain(self.output_texture)
    end

    self.configure_dirty = true
end

function BeautyService:SetExternalOutputTexture(texture)
    self.external_output_texture = texture

    self.configure_dirty = true
end

function BeautyService:_OnShutdown()
    if self.output_texture ~= nil then
        hl.Class_Release(self.output_texture)
        self.output_texture = nil
    end

    self.cur_beauty = nil

    hl.Class_Release(self.face_beauty_1)
    self.face_beauty_1_inited = false
    self.face_beauty_1 = nil

    hl.Class_Release(self.face_beauty_2)
    self.face_beauty_2_inited = false
    self.face_beauty_2 = nil

    hl.Class_Release(self.face_beauty_3)
    self.face_beauty_3_inited = false
    self.face_beauty_3 = nil
end

function BeautyService:GetOutputTexture()
    return self.external_output_texture or self.output_texture
end

return BeautyService