--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local GameFramework = import "GameFramework"
local NormalNodeLayer = import "NormalNodeLayer"

local EventRecordPlayNodeLayer = define ("EventRecordPlayNodeLayer", NormalNodeLayer)

LOGD("EventRecordPlayNodeLayer has been defined, parent is ", NormalNodeLayer)

function EventRecordPlayNodeLayer:_OnInitialize()
    self.message_processors = {
        [hl.Hl_MSG_SCRIPT_INITIALIZE_DEPRECATED] = { func = self.InitializeScript, can_record = true },

        [hl.Hl_MSG_SCRIPT_SHUTDOWN] = { func = self.ShutdownScript, can_record = true },

        [hl.Hl_MSG_SCRIPT_NODE_LOGIC] = { func = self.DispatchNodeEvent, can_record = false }
    }
end

function EventRecordPlayNodeLayer:InitializeScript(msg, message_source)
    local parent_name = msg:PopString()
    local parent_id = msg:PopInt32()

    local parent_node = self:QueryNode(parent_name, parent_id)
    if parent_node == nil then
        LOGE("InitializeScript: parent_name:", parent_name, "parent_id:",
                parent_id, "not found !!!")
        return false
    end

    local num_node = msg:PopUint32()
    LOGD("InitializeScript: parent_name:", parent_name, "parent_id:",
            parent_id, "num_node:", num_node, "message_source", message_source)
    local node_configs = {}
    local ok = true
    local node = nil
    for i = 1, num_node do
        local node_type = msg:PopString()
        local node_id = msg:PopInt32()

        node = self:QueryNode(node_type, node_id)

        if node ~= nil then
            ok = false
            LOGE("InitializeScript: node_name:", node_type, "node_id:",
                    node_id, "already exist, can not add it again !!!")
            break
        end

        local node_index = msg:PopInt32()
        local pos_x = msg:PopFloat()
        local pos_y = msg:PopFloat()
        local pos_z = msg:PopFloat()
        local rot_x = msg:PopFloat()
        local rot_y = msg:PopFloat()
        local rot_z = msg:PopFloat()
        local scl_x = msg:PopFloat()
        local scl_y = msg:PopFloat()
        local scl_z = msg:PopFloat()
        local info = {
            name = node_type,
            id = node_id,
            index = node_index,
            pos = hl.Vec3(pos_x, pos_y, pos_z),
            rot = hl.Quat(hl.Vec3(rot_x, rot_y, rot_z)),
            scl = hl.Vec3(scl_x, scl_y, scl_z),
        }
        table.insert(node_configs, info)
        LOGD("NodeManager: InitializeScript: Require: ", node_id, node_type,
                node_index, info.pos:CStr(), info.rot:CStr(), info.scl:CStr(), "Success")
    end

    if ok then
        ok = self:AddNodeWithConfig(node_configs, parent_name, parent_id)
        if false == ok then
            LOGE("NodeManager: InitializeScript: AddNodeWithConfig fail ...")
        end
    end

    return ok
end

function EventRecordPlayNodeLayer:AddNodeWithConfig(node_configs, parent_type, parent_id)
    local add_success = false
    local added_node = nil
    local added_nodes = {}

    for _, node_info in ipairs(node_configs) do
        CHECK(node_info.name ~= hl.Hl_SCRIPT_ROOT_NAME
                and node_info.id ~= hl.Hl_SCRIPT_ROOT_ID,
                "node name/id is invalid!!!", node_info.name, node_info.id, " use root name/id")
        add_success, added_node = self:AddNode(node_info.name, node_info.id)

        if not add_success then
            break
        else
            table.insert(added_nodes, added_node)
        end
    end

    if not add_success then
        for _, node in added_nodes do
            self:RemoveNode(node:GetType(), node:GetId())
        end
    end

    self.added_nodes = {}

    return add_success
end

function EventRecordPlayNodeLayer:ShutdownScript(msg, message_source)
    local node_type = msg:PopString()
    local node_id = msg:PopInt32()
    LOGD("NodeManager: ShutdownScript", node_type, node_id)

    return self:RemoveNode(node_type, node_id)
end

function EventRecordPlayNodeLayer:DispatchNodeEvent(msg, message_source)
    local node_type = msg:PopString()
    local node_id = msg:PopInt32()

    local node = self:QueryNode(node_type, node_id)
    if node == nil then
        return false
    end

    return node:NodeEvent(msg, message_source)
end

function EventRecordPlayNodeLayer:OnRecordedEvent(msg)
    local message_id = msg:PopBegin()

    local processor = self.message_processors[message_id]
    if processor == nil then
        CHECK(false, "EventRecordPlayNodeLayer: OnRecordedEvent: msg:", message_id, "has no processor !!!")
        return
    end

    processor.func(self, msg, GameFramework.KMessage_Recorded)
end