--==================================================================================
--Created:            2019.4.29
--Author:             Panhw
--==================================================================================
local Node = import "Node"
local Live2dAvatar = define("Live2dAvatar", Node)
local GameFramework = import "GameFramework"
local TextView = import "TextView"
local Service = import "Service"

local model_y_trans = 0.042

function Live2dAvatar:_OnInitialize()
    if hl.IsPlatformDesktop() then
        GameFramework:AddListener(hl.Hl_MSG_MOUSE, self.MouseEvent, self)
    else
        GameFramework:AddListener(hl.Hl_MSG_TOUCH, self.TouchEvent, self)
    end

    if hl.IsPlatformiOS() then
        self:GetService(Service.SERVICE_SHOW_BG):EnableBG(self.node_context.node_manager.INPUT_TEXTURE_0, true)
    end
    

    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_MODEL_CREATE, self.OnModelCreate, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_LOAD_PHYSICS, self.OnLoadPhysics, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_LOAD_POSE, self.OnLoadPose, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_LOAD_TEXTURE, self.OnLoadTexture, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_LOAD_EXPRESSION, self.OnLoadExpression, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_LOAD_MOTION, self.OnLoadMotion, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_START_MOTION, self.OnStartMotion, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_STOP_MOTION, self.OnStopMotion, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_SET_EXPRESSTION, self.OnSetExpression, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_REPLACE_TEXTURE, self.OnReplaceTexture, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_REMOVE_TEXTURE, self.OnRemoveTexture, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_SET_PARAMETER, self.OnSetParameter, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_SET_ACTIONS, self.OnSetActions, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_SET_TEXT, self.OnSetText, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_SET_BACKGROUND, self.OnSetBackground, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_SET_MODE, self.OnSetMode, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_SET_STYLE, self.OnSetStyle, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_MODEL_RELEASE, self.OnModelRelease, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_SET_MODEL_VISIBILE, self.OnSetModelVisibile, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_SET_MOTION_STREAK, self.OnSetMotionStreak, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_SET_MODEL_SLOT, self.OnSetModelSlot, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_SET_ENABLE_BREATH, self.OnSetEnableBreath, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_SET_ENABLE_EYEBLINK, self.OnSetEnableEyeBlink, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_ENABLE_AVATAR_TRACKER, self.OnSetEnableAvatarTracker, self)
    GameFramework:AddListener(hl.Hl_MSG_LIVE2D_ENABLE_WATERMARK, self.OnSetEnableWatermark, self)

    local world = GameFramework:GetNodeManager():GetWorld()
    world:CreateSystem(hl.AnimatorSystem_ID())
    world:CreateSystem(hl.Live2DSystem_ID())

    self.work_dir = hl.GetWorkDir()
    self.live2d_models = {}

    local c2d = GameFramework:GetNodeManager():GetDefaultCamera2d()
    local bg_ac = hl.Actor_New()
    bg_ac:SetName(hl.NamedView("Live2D_BgActor"))

    local bg_sp = hl.Sprite_New()
    local bg_mtl = hl.Sprite_CreateSimpleMaterial()
    local bg_rs = hl.GPRenderState()
    bg_rs.blend_src = hl.GPBlendMode_kSrcAlpha
    bg_rs.blend_dst = hl.GPBlendMode_kSrcAlphaInv
    bg_mtl:SetRenderState(bg_rs)
    bg_sp:SetMaterial(bg_mtl)
    bg_sp:SetVisible(false)
    bg_sp:SetLayer(800)
    bg_sp:SetFilterMask(c2d:GetFilterMask())
    bg_ac:AddComponent(bg_sp)
    self:AddToWorld(bg_ac)
    self.bg_ac = bg_ac
    self.bg_sp = bg_sp

    local watermark_ac = hl.Actor_New()
    watermark_ac:SetName(hl.NamedView("Live2D_WatermarkActor"))
    local watermark_sp = hl.Sprite_New()
    local watermark_mtl = hl.Sprite_CreateSimpleMaterial(self.work_dir / "Live2dAvatar/Asset/watermark.png", true)
    local watermark_rs = hl.GPRenderState()
    watermark_rs.blend_src = hl.GPBlendMode_kSrcAlpha
    watermark_rs.blend_dst = hl.GPBlendMode_kSrcAlphaInv
    watermark_mtl:SetRenderState(watermark_rs)
    watermark_sp:SetMaterial(watermark_mtl)
    watermark_sp:SetVisible(false)
    watermark_sp:SetLayer(10000)
    local watermark_tex = watermark_mtl:GetTexture(0)
    watermark_sp:SetWidth(watermark_tex:GetWidth())
    watermark_sp:SetHeight(watermark_tex:GetHeight())
    watermark_sp:SetFilterMask(c2d:GetFilterMask())
    watermark_ac:AddComponent(watermark_sp)
    self:AddToWorld(watermark_ac)
    self.watermark_ac = watermark_ac
    self.watermark_sp = watermark_sp

    local motion_streak_rate = hl.RuntimeWidth() / 360
    local motion_streak_ac = hl.Actor_New()
    motion_streak_ac:SetName(hl.NamedView("Live2D_MotionStreak"))
    local motion_streak = hl.MotionStreak_New()
    local motion_streak_mtl = hl.MotionStreak_MakeMaterial()
    local motion_streak_rs = motion_streak_mtl:GetRenderState()
    motion_streak_rs.blend_dst = hl.GPBlendMode_kSrcAlphaInv
    motion_streak_mtl:SetRenderState(motion_streak_rs)
    motion_streak:SetMaterial(motion_streak_mtl)
    motion_streak:SetFadeTime(500)
    -- motion_streak:SetColor(0xF6C1C6FF)
    motion_streak:SetColor(0xFFFFFFFF)
    motion_streak:SetWidth(motion_streak_rate * 11)
    motion_streak:SetLayer(10000)
    motion_streak:SetFilterMask(c2d:GetFilterMask())
    motion_streak_ac:AddComponent(motion_streak)
    self:AddToWorld(motion_streak_ac)
    self.motion_streak_ac = motion_streak_ac
    self.motion_streak = motion_streak

    local pe = hl.SphereEmitter_New()
    pe:SetEnable(false)
    -- local particle_palette = hl.U32Array()
    -- particle_palette:push_back(0xFFBEBEFF)
    -- particle_palette:push_back(0xFFBEF2FF)
    -- particle_palette:push_back(0xE3BEFFFF)
    -- particle_palette:push_back(0xBECAFFFF)
    -- particle_palette:push_back(0xBEFBFFFF)
    -- particle_palette:push_back(0xBEFFCDFF)
    -- particle_palette:push_back(0xF0FFBEFF)
    -- particle_palette:push_back(0xFFD8BEFF)
    -- pe:SetStartPalette(particle_palette)
    hl.ParticleEmitter_ConfigToDefault(pe)
    local v = hl.RandomValue_New()
    v:SetMinMax(hl.Vec2(motion_streak_rate * 20.0, motion_streak_rate * 60.0))
    pe:SetStartSize(v)
    v = hl.RandomValue_New()
    v:SetMinMax(hl.Vec2(45.0, 180.0))
    pe:SetStartSpeed(v)
    v = hl.RandomValue_New()
    v:SetMinMax(hl.Vec2(1200, 2000.0))
    pe:SetStartLifetime(v)
    v = hl.RandomValue_New()
    v:SetMinMax(hl.Vec2(14.0, 35.0))
    pe:SetEmissionRate(v)
    local ps = hl.Particles_New()
    ps:SetFilterMask(c2d:GetFilterMask())
    ps:SetLayer(10000)
    ps:SetMaxParticleCount(1000)
    ps:SetUseWorldSpace(true)
    ps:SetSortMode(hl.Particles_kByDistance)
    ps:SetEmitter(pe)
    local color_affector = hl.ColorAffector_New()
    local hue = hl.TweenValue_New()
    hue:SetBeginEnd(hl.Vec2(0, 0))
    color_affector:SetHue(hue)
    local saturation = hl.TweenValue_New()
    saturation:SetBeginEnd(hl.Vec2(0, 0))
    color_affector:SetSaturation(saturation)
    local value = hl.TweenValue_New()
    value:SetBeginEnd(hl.Vec2(1, 1))
    color_affector:SetValue(value)
    local alpha = hl.TweenValue_New()
    alpha:SetBeginEnd(hl.Vec2(1, 0))
    color_affector:SetAlpha(alpha)
    ps:AddAffector(color_affector)

    local uv_affector = hl.UVAnimAffector_New()
    -- uv_affector:SetTileColumn(6)
    local fot = hl.ConstantValue_New()
    fot:SetValue(0)
    uv_affector:SetFrameOverTime(fot)
    uv_affector:SetStartMinFrame(0)
    uv_affector:SetStartMaxFrame(5)
    ps:AddAffector(uv_affector)
    local ps_mtl = hl.Particles_CreateSimpleMaterial(hl.Particles_kAdd)
    self.tex = ps_mtl:GetTexture(0)
    ps:SetMaterial(ps_mtl)
    ps:SetPlayOnAwake(true)

    local ps_ac = hl.Actor_New()
    ps_ac:SetName(hl.NamedView("Live2D_Particle"))
    ps_ac:AddComponent(ps)
    self:AddToWorld(ps_ac)
    self.particle_ac = ps_ac
    self.particle = ps
    self.uv_affector = uv_affector
    self.particle_emitter = pe

    self.avatar_tracker = hl.AvatarTracker_New()

    if hl.IsEditorMode() then
        -- self:InitTestModel(1, "Live2D", "Live2Ds/Test")
        -- self:InitTestModel(2, "Live2D/Body0", "Live2D/Body0")
        self:InitTestModel(1, "Live2D/Body1", "Live2D/Body1")
        -- self:InitTestModel(3, "Live2D/Hair0", "Live2D/Hair0")
        self:InitTestModel(2, "Live2D/Head", "Live2D/Head")
        self:InitTestModel(3, "Live2D/Hair1", "Live2D/Hair1")
        -- self:InitTestModelNew()
        self:OnSetModelSlot(1, hl.Hl_LIVE2D_SLOT_RIGHT)
        self:OnSetModelSlot(2, hl.Hl_LIVE2D_SLOT_RIGHT)--hl.Hl_LIVE2D_SLOT_DEFAULT)
        self:OnSetModelSlot(3, hl.Hl_LIVE2D_SLOT_RIGHT)
        
        self:InitTestModel(4, "Live2D/Boy", "Live2D/Boy")
        self:OnSetModelSlot(4, hl.Hl_LIVE2D_SLOT_LEFT)

        self:OnSetMotionStreak(-1, true, (self.work_dir / "Live2dAvatar/Asset/ParticleConfig.json"):CStr())
        self:OnSetEnableWatermark(-1, true)
    end
    -- self:OnSetEnableFaceTracker(1, true)
    
    --self:_InitDefaultBlackBG()
    return true
end

function Live2dAvatar:_InitDefaultBlackBG()
   if hl.IsPlatformAndroid() then
        local c2d = GameFramework:GetNodeManager():GetDefaultCamera2d()
        local bg_ac = hl.Actor_New()
        bg_ac:SetName(hl.NamedView("Live2D_DefaultBlackBgActor"))
        local bg_sp = hl.Sprite_New()
        local bg_mtl = hl.Sprite_CreateSimpleMaterial()
        local bg_rs = hl.GPRenderState()
        bg_rs.blend_src = hl.GPBlendMode_kSrcAlpha
        bg_rs.blend_dst = hl.GPBlendMode_kSrcAlphaInv
        bg_mtl:SetRenderState(bg_rs)
        bg_sp:SetMaterial(bg_mtl)
        bg_sp:SetVisible(false)
        bg_sp:SetLayer(700)
        bg_sp:SetFilterMask(c2d:GetFilterMask())
        bg_sp:GetMaterial():SetTexture(0, self:GetDefaultBlackTexture())
        bg_sp:SetWidth(hl.RuntimeWidth())
        bg_sp:SetHeight(hl.RuntimeHeight())
        bg_sp:SetVisible(true)
        bg_ac:AddComponent(bg_sp)
        self:AddToWorld(bg_ac)
    end
end

function Live2dAvatar:_OnShutdown()
    if hl.IsPlatformiOS() then
        self:GetService(Service.SERVICE_SHOW_BG):EnableBG(self.node_context.node_manager.INPUT_TEXTURE_0, false)
    end

    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_MODEL_CREATE, self.OnModelCreate)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_LOAD_PHYSICS, self.OnLoadPhysics)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_LOAD_POSE, self.OnLoadPose)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_LOAD_TEXTURE, self.OnLoadTexture)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_LOAD_EXPRESSION, self.OnLoadExpression)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_LOAD_MOTION, self.OnLoadMotion)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_START_MOTION, self.OnStartMotion)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_STOP_MOTION, self.OnStopMotion)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_SET_EXPRESSTION, self.OnSetExpression)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_REPLACE_TEXTURE, self.OnReplaceTexture)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_REMOVE_TEXTURE, self.OnRemoveTexture)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_SET_PARAMETER, self.OnSetParameter)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_SET_ACTIONS, self.OnSetActions)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_SET_TEXT, self.OnSetText)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_SET_BACKGROUND, self.OnSetBackground)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_SET_MODE, self.OnSetMode)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_SET_STYLE, self.OnSetStyle)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_MODEL_RELEASE, self.OnModelRelease)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_SET_MODEL_VISIBILE, self.OnSetModelVisibile)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_SET_MOTION_STREAK, self.OnSetMotionStreak)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_SET_MODEL_SLOT, self.OnSetModelSlot)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_SET_ENABLE_BREATH, self.OnSetEnableBreath)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_SET_ENABLE_EYEBLINK, self.OnSetEnableEyeBlink)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_ENABLE_AVATAR_TRACKER, self.OnSetEnableAvatarTracker)
    GameFramework:RemoveListener(hl.Hl_MSG_LIVE2D_ENABLE_WATERMARK, self.OnSetEnableWatermark)

    if hl.IsPlatformDesktop() then
        GameFramework:RemoveListener(hl.Hl_MSG_MOUSE, self.MouseEvent)
    else
        GameFramework:RemoveListener(hl.Hl_MSG_TOUCH, self.TouchEvent)
    end
    for k, v in pairs(self.live2d_models) do
        self:OnModelRelease(k)
    end
    if self.bg_ac then
        self.node_context.node_manager:GetWorld():RemoveActor(self.bg_ac)
        self.bg_ac = nil
    end
    if self.watermark_ac then
        self.node_context.node_manager:GetWorld():RemoveActor(self.watermark_ac)
        self.watermark_ac = nil
    end
    if self.motion_streak_ac then
        self.node_context.node_manager:GetWorld():RemoveActor(self.motion_streak_ac)
        self.motion_streak_ac = nil
    end
    if self.particle_ac then
        self.node_context.node_manager:GetWorld():RemoveActor(self.particle_ac)
        self.particle_ac = nil
    end
    if self.avatar_tracker then
        hl.Class_Discard(self.avatar_tracker)
        self.avatar_tracker = nil
    end
end

function Live2dAvatar:_PlayText(user_id, time)
    local model = self.live2d_models[user_id]
    if model and model.visibile == 1 then
        local fade_time = time * 0.2
        if fade_time > 800 then
            fade_time = 800
        end
        model.text_sp:GetMaterial():SetAlpha(0)
        model.textbg_sp:GetMaterial():SetAlpha(0)
        local anim_channel = hl.MaterialVec4AnimChannel_Cast(model.text_anim_src:GetChannel(0))
        anim_channel.property_times[0] = 0
        anim_channel.property_times[1] = math.min(fade_time * 0.1, 120)
        anim_channel.property_times[2] = fade_time
        anim_channel.property_times[3] = time - fade_time
        anim_channel.property_times[4] = time * 0.99
        anim_channel = hl.MaterialVec4AnimChannel_Cast(model.text_anim_src:GetChannel(1))
        anim_channel.property_times[1] = 0
        anim_channel.property_times[1] = math.min(fade_time * 0.1, 120)
        anim_channel.property_times[2] = fade_time
        anim_channel.property_times[3] = time - fade_time
        anim_channel.property_times[4] = time * 0.99
        model.text_anim_src:SetDuration(time)
        model.text_anim:RemoveAllClips()
        model.text_anim:AddClip(model.text_anim_src:NewClip())
        if model.text_anim:CollectTargets(false) then
            model.text_anim:Play()
        end
    end
end

function Live2dAvatar:_SetText(user_id, show, text)
    local model = self.live2d_models[user_id]
    if model and model.visibile == 1 then
        if show == 1 and text ~= "" then
            local tex = model.sys_text:UpdateText(text)
            if tex then
                local offset = -(self.text_x_offset - tex:GetWidth() / 2.0)
                model.text_ac:GetTransform():SetLocalPosition(hl.Vec3(offset, -self.text_bg_y_offset, 0))
                model.text_sp:GetMaterial():SetAlpha(1)
                model.text_sp:GetMaterial():SetTexture(0, tex)
                model.text_sp:SetWidth(tex:GetWidth())
                model.text_sp:SetHeight(tex:GetHeight())
                model.text_sp:SetVisible(true)

                model.textbg_sp:GetMaterial():SetAlpha(1)
                model.textbg_sp:SetWidth(self.text_bg_width)
                local h = self.text_bg_min_height
                if self.text_bg_min_height < tex:GetHeight() + 2 * self.text_magin then
                    h =  tex:GetHeight() + 2 * self.text_magin
                end
                model.textbg_sp:SetHeight(h)
                model.textbg_sp:SetVisible(true)
            else
                model.text_sp:SetVisible(false)
                model.textbg_sp:SetVisible(false)
            end
        else
            model.text_sp:SetVisible(false)
            model.textbg_sp:SetVisible(false)
        end
    end
end

function Live2dAvatar:_ApplyTrigger(user_id, live2d, trigger)
    local model = self.live2d_models[user_id]
    if model and model.visibile == 1 then
        if trigger.type == "text" then
            local index = 1
            if trigger.method == "random" then
                index = math.random(1, #trigger.value)
            elseif trigger.method == "forward" then
                trigger.__script_index = trigger.__script_index or 0
                trigger.__script_index = trigger.__script_index + 1
                if trigger.__script_index > #trigger.value then
                    trigger.__script_index = 1
                end
                index = trigger.__script_index
            end
            local text = trigger.value[index]
            local time_sec = 2
            if #trigger.times == #trigger.value then
                time_sec = trigger.times[index]
            elseif #trigger.times > 0 then
                time_sec = trigger.times[1]
            end
            if model.mode == 1 then
                local send_msg = hl.MessageQueue_MakeAsync()
                send_msg:PushBegin(hl.Hl_MSG_NATIVE_LIVE2D_SHOW_TEXT)
                send_msg:PushInt32(user_id)
                send_msg:PushString(text)
                send_msg:PushFloat(time_sec)
                send_msg:PushEnd()
                hl.MessageQueue_PostAndFreeAsync(send_msg)
            else
                self:_SetText(user_id, 1, text)
                self:_PlayText(user_id, time_sec * 1000)
            end
        elseif trigger.type == "motions" then
            if #trigger.value > 0 then
                local index = 1
                if trigger.method == "random" then
                    index = math.random(1, #trigger.value)
                elseif trigger.method == "forward" then
                    trigger.__script_index = trigger.__script_index or 0
                    trigger.__script_index = trigger.__script_index + 1
                    if trigger.__script_index > #trigger.value then
                        trigger.__script_index = 1
                    end
                    index = trigger.__script_index
                end
                local name = trigger.value[index]
                local pa = hl.PathArray()
                pa:push_back(hl.Path(name))
                live2d:StartMotion(pa)
                LOGD("StartMotion:", name)
            end
        elseif trigger.type == "expressions" then
            if #trigger.value > 0 then
                local index = 1
                if trigger.method == "random" then
                    index = math.random(1, #trigger.value)
                elseif trigger.method == "forward" then
                    trigger.__script_index = trigger.__script_index or 0
                    trigger.__script_index = trigger.__script_index + 1
                    if trigger.__script_index > #trigger.value then
                        trigger.__script_index = 1
                    end
                    index = trigger.__script_index
                end
                live2d:SetExpression(hl.Path(trigger.value[index]))
            end
        elseif trigger.type == "components" then
            local a = #trigger.value
            local b = #trigger.replace
            if a == b and a > 0 then
                for k, v in ipairs(trigger.value) do
                    local name = trigger.value[k]
                    local replace = trigger.replace[k]
                    live2d:ReplaceTexture(name, hl.Path(replace))
                end
            end
        elseif trigger.type == "voice" then

        end
    end
end
function Live2dAvatar:_OnAwake()

end

function Live2dAvatar:_OnSleep()

end

function Live2dAvatar:TouchEvent(action, touches)
    if action == hl.Hl_ACTION_TOUCH_BEGAN then
		self:_ClickBegan(touches[1].x, touches[1].y)
	elseif action == hl.Hl_ACTION_TOUCH_MOVED then
		self:_ClickMoved(touches[1].x, touches[1].y)
	elseif action == hl.Hl_ACTION_TOUCH_ENDED then
		self:_ClickEnded(touches[1].x, touches[1].y)
    elseif action == hl.Hl_ACTION_TOUCH_CANCELLED then
        self:_ClickEnded(touches[1].x, touches[1].y)
	end
end

function Live2dAvatar:MouseEvent(action, x, y)
    if action == hl.Hl_ACTION_MOUSE_ENTER_LEFT then
		self:_ClickBegan(x, y)
	elseif action == hl.Hl_ACTION_MOUSE_MOVED then
		self:_ClickMoved(x, y)
	elseif action == hl.Hl_ACTION_MOUSE_LEAVE_LEFT then
		self:_ClickEnded(x, y)
	end
end

function Live2dAvatar:_GetModelNum()
    local num = 0
    for _, model in pairs(self.live2d_models) do
        if model ~= nil then
            num = num + 1
        end
    end

    return num
end

function Live2dAvatar:_ClickBegan(x, y)
    if self:_GetModelNum() > 0 then
        local camera = GameFramework:GetNodeManager():GetDefaultCamera2d()
        local ray2d = camera:ScreenToRay(hl.Vec2(x, y))
        local point = hl.Vec2(0)
        for k, v in pairs(self.live2d_models) do
            if v.visibile == 1 then
                if v.sp:RayCastNDCPoint(ray2d, point) then
                    self:_HitTest(k, 1, point.x, point.y)
                    for _, action in ipairs(v.cfg.actions) do
                        -- if v.l2d:HitTest(action.hitArea, point.x, point.y) then
                            for _, trigger in ipairs(action.triggers) do
                                self:_ApplyTrigger(k, v.l2d, trigger)
                            end
                        -- end
                    end
                end
            end
        end
    end
    if self.enable_motion_streak then
        if not self.motion_streak_click then
            self.motion_streak_click = true
            self.motion_streak:Begin()
            local c2d = GameFramework:GetNodeManager():GetDefaultCamera2d()
            local pos = c2d:ScreenToWorld(hl.Vec3(x, y, 0))
            pos.z = 0
            self.motion_streak_ac:GetTransform():SetGlobalPosition(pos)
            self.particle_ac:GetTransform():SetGlobalPosition(pos)
            self.particle_emitter:SetEnable(true)
        end
    end
end 

function Live2dAvatar:_ClickMoved(x, y)
    if self:_GetModelNum() > 0 then
        local c2d = GameFramework:GetNodeManager():GetDefaultCamera2d()
        local ray2d = c2d:ScreenToRay(hl.Vec2(x, y))
        local point = hl.Vec2(0)
        for _, v in pairs(self.live2d_models) do
            if v.visibile == 1 then
                if v.sp:RayCastNDCPoint(ray2d, point) then
                    v.l2d:SetDragging(point.x, point.y)
                else
                    v.l2d:SetDragging(0, 0)
                end
            end
        end
    end
    if self.enable_motion_streak then
        if self.motion_streak_click then
            local c2d = GameFramework:GetNodeManager():GetDefaultCamera2d()
            local pos = c2d:ScreenToWorld(hl.Vec3(x, y, 0))
            pos.z = 0
            self.motion_streak_ac:GetTransform():SetGlobalPosition(pos)
            self.particle_ac:GetTransform():SetGlobalPosition(pos)
        end
    end
end

function Live2dAvatar:_ClickEnded(x, y)
    local camera = GameFramework:GetNodeManager():GetDefaultCamera2d()
    local ray2d = camera:ScreenToRay(hl.Vec2(x, y))
    local point = hl.Vec2(0)
    for k, v in pairs(self.live2d_models) do
        if v.visibile == 1 then
            v.l2d:SetDragging(0, 0)
            if v.sp:RayCastNDCPoint(ray2d, point) then
                self:_HitTest(k, 2, point.x, point.y)
            end
        end
    end
    if self.enable_motion_streak then
        if self.motion_streak_click then
            self.motion_streak_click = false
            local c2d = GameFramework:GetNodeManager():GetDefaultCamera2d()
            local pos = c2d:ScreenToWorld(hl.Vec3(x, y, 0))
            pos.z = 0
            self.motion_streak_ac:GetTransform():SetGlobalPosition(pos)
            self.particle_ac:GetTransform():SetGlobalPosition(pos)
            self.motion_streak:End()
            self.particle_emitter:SetEnable(false)
        end
    end
end

function Live2dAvatar:_HitTest(user_id, event, x, y)
    local model = self.live2d_models[user_id]
    local hit = false
    if model.cfg.touchAreas then
        for _, touch_area in ipairs(model.cfg.touchAreas) do
            if model.l2d:HitTest(touch_area, x, y) then
                local send_msg = hl.MessageQueue_MakeAsync()
                send_msg:PushBegin(hl.Hl_MSG_NATIVE_LIVE2D_TOUCH_EVENT)
                send_msg:PushInt32(user_id)
                send_msg:PushString(touch_area)
                send_msg:PushInt32(event)
                send_msg:PushEnd()
                hl.MessageQueue_PostAndFreeAsync(send_msg)
                hit = true
            end
        end
    end
    return hit
end  

function Live2dAvatar:_OnUpdate(dt)
    self:_UpdateFaceTracker()
    self:_UpdateAvatarTracker()
end

function Live2dAvatar:OnModelCreate(user_id, config)
    local cfg = cjson.decode(config)
    self:OnModelInitWithJson(user_id, cfg)
end

function Live2dAvatar:OnModelInitWithJson(user_id, cfg)
    local l2d_all_tex_ids = hl.StrArray()
    local l2d_motions = hl.PathArray()
    local l2d_expressions = hl.PathArray()
    local l2d_ids = hl.StrArray()
    local l2d_textures = hl.PathArray()
    for _, v in ipairs(cfg.componentTypes) do
        l2d_all_tex_ids:push_back(v)
    end
    for _, v in ipairs(cfg.motions) do
        l2d_motions:push_back(hl.Path(v.motionId))
    end
    for _, v in ipairs(cfg.expressions) do
        l2d_expressions:push_back(hl.Path(v.expressionId))
    end
    for _, v in ipairs(cfg.components) do
        l2d_ids:push_back(v.type)
        l2d_textures:push_back(hl.Path(v.componentId))
    end
    local sp_width, sp_height
    local local_moc_scale = 2660 / 2160
    sp_width = hl.RuntimeWidth() * (cfg.modelMocScale or local_moc_scale)
    self.text_size = 16.0 / 375.0 * hl.RuntimeWidth()
    self.text_magin = 20 / 375.0 * hl.RuntimeWidth()
    self.text_bg_width = hl.RuntimeWidth() - 2 * 10 / 375.0 * hl.RuntimeWidth()
    self.text_bg_min_height = 110.0 / 375.0 * hl.RuntimeWidth()
    self.text_bg_y_offset= hl.RuntimeHeight() / 2.0 - 180.0 / 667.0 * hl.RuntimeHeight()
    self.text_x_offset = self.text_bg_width / 2.0 -  self.text_magin

    local live2d = hl.Live2DPlayer_New()
    if live2d:InitWithSource(hl.Path(cfg.avatarId), l2d_all_tex_ids, cfg.num_motion_queue or 6) then
        if live2d:LoadMotions(l2d_motions)
            and live2d:LoadExpressions(l2d_expressions)
            and live2d:LoadTextures(l2d_ids, l2d_textures) then
            if cfg.pose and cfg.pose.poseId then
                live2d:LoadPose(hl.Path(cfg.pose.poseId))
            end
            if cfg.physics and cfg.physics.physicsId then
                live2d:LoadPhysics(hl.Path(cfg.physics.physicsId))
            end
            live2d:SetMatrixScale(cfg.modelMatScale or 3.45)
            local c2d = GameFramework:GetNodeManager():GetDefaultCamera2d()
            local l2d_root = hl.Actor_New()
            l2d_root:SetName(hl.NamedView("Live2DRootActor" .. user_id))
            local ac = hl.Actor_New()
            ac:SetName(hl.NamedView("Live2DActor" .. user_id))
            local sp = hl.Sprite_New()
            local render_state = hl.GPRenderState()
            render_state.blend_src = hl.GPBlendMode_kSrcAlpha
            render_state.blend_dst = hl.GPBlendMode_kSrcAlphaInv
            local material = live2d:GetMaterial()
            material:SetRenderState(render_state)
            sp:SetMaterial(material)
            sp_height = sp_width * live2d:GetPixelHeight() / live2d:GetPixelWidth()
            sp:SetWidth(sp_width)
            sp:SetHeight(sp_height)
            sp:SetLayer(1000)
            sp:SetFilterMask(c2d:GetFilterMask())
            ac:AddComponent(live2d)
            ac:AddComponent(sp)
            ac:GetTransform():SetLocalPosition(hl.Vec3(0, (sp_height - hl.RuntimeHeight()) * 0.5 - hl.RuntimeHeight() * model_y_trans * local_moc_scale, 0))
            ac:GetTransform():SetParent(l2d_root:GetTransform())
            for _, v in ipairs(cfg.components) do
                if v.parameters then
                    for k, p in ipairs(v.parameters) do
                        live2d:SetParameter(p.parameterId, p.value)
                    end
                end
            end

            local bg_ac = hl.Actor_New()
            bg_ac:SetName(hl.NamedView("Live2D_BgActor" .. user_id))
            local bg_sp = hl.Sprite_New()
            local bg_mtl = hl.Sprite_CreateSimpleMaterial()
            local bg_rs = hl.GPRenderState()
            bg_rs.blend_src = hl.GPBlendMode_kSrcAlpha
            bg_rs.blend_dst = hl.GPBlendMode_kSrcAlphaInv
            bg_mtl:SetRenderState(bg_rs)
            bg_sp:SetMaterial(bg_mtl)
            bg_sp:SetVisible(false)
            bg_sp:SetLayer(900)
            bg_sp:SetFilterMask(c2d:GetFilterMask())
            bg_ac:AddComponent(bg_sp)
            bg_ac:GetTransform():SetParent(l2d_root:GetTransform())
            local textbg_ac = hl.Actor_New()
            textbg_ac:SetName(hl.NamedView("Live2D_TextBgActor" .. user_id))
            local textbg_sp = hl.Sprite_New()
            local textbg_mtl = hl.Sprite_CreateFlipMaterial()
            local textbg_rs = hl.GPRenderState()
            textbg_rs.blend_src = hl.GPBlendMode_kSrcAlpha
            textbg_rs.blend_dst = hl.GPBlendMode_kSrcAlphaInv
            textbg_mtl:SetRenderState(textbg_rs)
            textbg_sp:SetMaterial(textbg_mtl)
            textbg_sp:SetVisible(false)
            textbg_sp:SetLayer(1090)
            textbg_sp:SetFilterMask(c2d:GetFilterMask())
            textbg_ac:AddComponent(textbg_sp)
            textbg_ac:GetTransform():SetParent(l2d_root:GetTransform())
            local text_ac = hl.Actor_New()
            text_ac:SetName(hl.NamedView("Live2D_TextActor" .. user_id))
            local text_sp = hl.Sprite_New()
            local text_mtl = hl.Sprite_CreateFlipMaterial()
            local text_rs = hl.GPRenderState()
            text_rs.blend_src = hl.GPBlendMode_kSrcAlpha
            text_rs.blend_dst = hl.GPBlendMode_kSrcAlphaInv
            text_mtl:SetRenderState(text_rs)
            text_sp:SetMaterial(text_mtl)
            text_sp:SetVisible(false)
            text_sp:SetLayer(1100)
            text_sp:SetFilterMask(c2d:GetFilterMask())
            text_ac:AddComponent(text_sp)
            text_ac:GetTransform():SetParent(l2d_root:GetTransform())

            local animation = hl.Animation_New()
            local anim_channel = hl.MaterialVec4AnimChannel_New()
            anim_channel:SetName(hl.NamedView("Live2D_TextActor" .. user_id))
            anim_channel.property_name = hl.Named("u_Setting")
            anim_channel.property_values:push_back(hl.Vec4(0))
            anim_channel.property_values:push_back(hl.Vec4(0))
            anim_channel.property_values:push_back(hl.Vec4(1))
            anim_channel.property_values:push_back(hl.Vec4(1))
            anim_channel.property_values:push_back(hl.Vec4(0))
            anim_channel.property_times:push_back(0)
            anim_channel.property_times:push_back(100)
            anim_channel.property_times:push_back(300)
            anim_channel.property_times:push_back(1200)
            anim_channel.property_times:push_back(1500)
            animation:AddChannel(anim_channel)

            anim_channel = hl.MaterialVec4AnimChannel_New()
            anim_channel:SetName(hl.NamedView("Live2D_TextBgActor" .. user_id))
            anim_channel.property_name = hl.Named("u_Setting")
            anim_channel.property_values:push_back(hl.Vec4(0))
            anim_channel.property_values:push_back(hl.Vec4(0))
            anim_channel.property_values:push_back(hl.Vec4(1))
            anim_channel.property_values:push_back(hl.Vec4(1))
            anim_channel.property_values:push_back(hl.Vec4(0))
            anim_channel.property_times:push_back(0)
            anim_channel.property_times:push_back(100)
            anim_channel.property_times:push_back(300)
            anim_channel.property_times:push_back(1200)
            anim_channel.property_times:push_back(1500)
            animation:AddChannel(anim_channel)

            animation:SetDuration(1500)
            hl.Class_Retain(animation)
            local text_anim = hl.Animator_New()
            text_anim:SetPlayOnAwake(true)
            l2d_root:AddComponent(text_anim)

            local textbg_tex = hl.ImageTexture_New()
            if textbg_tex:InitWithSource(self.work_dir / "Live2dAvatar/Asset/text_bg.png") then
                textbg_mtl:SetTexture(0, textbg_tex)
            else
                hl.Class_Discard(tex)
            end
            textbg_ac:GetTransform():SetLocalPosition(hl.Vec3(0, -self.text_bg_y_offset, 0))
            local sys_text = hl.SysText_New()
            sys_text.width =  self.text_bg_width -  2 * self.text_magin
            sys_text.size = self.text_size
            sys_text.color = 0x333333FF
            sys_text.horizontal_align = hl.SysText_kLeftOrTop
            hl.Class_Retain(sys_text)
            
            local root_group = hl.ActorGroup_New()
            hl.Class_Retain(root_group)
            root_group:Collect(l2d_root)
            self:AddToWorld(root_group)
            self.live2d_models[user_id] = {
                cfg = cfg,
                ac = ac,
                sp = sp,
                l2d = live2d,
                bg_ac = bg_ac,
                bg_sp = bg_sp,
                text_ac = text_ac,
                text_sp = text_sp,
                text_anim = text_anim,
                text_anim_src = animation,
                textbg_ac = textbg_ac,
                textbg_sp = textbg_sp,
                sys_text = sys_text,
                root_ac = l2d_root,
                root_group = root_group,
                mode = 0,
                style = 0,
                visibile = 1,
                mac_scale = local_moc_scale,
                parameters = {},
            }

            live2d:SetEnableBreath(true)
            live2d:SetEnableEyeBlink(true)
            return
        end
    end
    hl.Class_Discard(live2d)
end

function Live2dAvatar:OnLoadPhysics(user_id, path)
    local model = self.live2d_models[user_id]
    if model then
        model.l2d:LoadPhysics(path)
    end
end

function Live2dAvatar:OnLoadPose(user_id, path)
    local model = self.live2d_models[user_id]
    if model then
        model.l2d:LoadPose(path)
    end
end

function Live2dAvatar:OnLoadTexture(user_id, tex_ids, tex_paths)
    local model = self.live2d_models[user_id]
    if model then
        model.l2d:LoadTexture(tex_ids, tex_paths)
    end
end

function Live2dAvatar:OnLoadExpression(user_id, expressions)
    local model = self.live2d_models[user_id]
    if model then
        model.l2d:LoadExpressions(expressions)
    end
end

function Live2dAvatar:OnLoadMotion(user_id, motions)
    local model = self.live2d_models[user_id]
    if model then
        model.l2d:LoadMotions(motions)
    end
end

function Live2dAvatar:OnStartMotion(user_id, motions, queue_index, looper, force)
    local model = self.live2d_models[user_id]
    if model then
        force = force or true
        model.l2d:StartMotion(motions, queue_index, looper, force)
    end
end

function Live2dAvatar:OnStopMotion(user_id, stop_all, queue_index)
    local model = self.live2d_models[user_id]
    if model then
        if stop_all then
            model.l2d:StopAllMotions()
        else
            model.l2d:StopMotion(queue_index)
        end
    end
end

function Live2dAvatar:OnSetExpression(user_id, expression)
    local model = self.live2d_models[user_id]
    if model then
        model.l2d:SetExpression(expression)
    end
end

function Live2dAvatar:OnReplaceTexture(user_id, tex_ids, tex_paths)
    local model = self.live2d_models[user_id]
    if model then
        for i = 0, tex_ids:size() - 1 do
            model.l2d:ReplaceTexture(tex_ids[i], tex_paths[i])
        end
    end
end

function Live2dAvatar:OnRemoveTexture(user_id, tex_ids, tex_paths)
    local model = self.live2d_models[user_id]
    if model then
        for i = 0, tex_ids:size() - 1 do
            model.l2d:ReplaceTexture(tex_ids[i], hl.Path(""))
        end
    end
end

function Live2dAvatar:OnSetParameter(user_id, id, value)
    local model = self.live2d_models[user_id]
    if model then
        model.l2d:SetParameter(id, value)
    end
end

function Live2dAvatar:OnSetTempParameter(user_id, id, value)
    local model = self.live2d_models[user_id]
    if model then
        model.l2d:SetTempParameter(id, value)
    end
end

function Live2dAvatar:OnSetActions(user_id, json_cfg)
    local model = self.live2d_models[user_id]
    if model then
        local cfg = cjson.decode(json_cfg)
        if cfg and cfg.actions then
            model.cfg.actions = cfg.actions
        end
    end
end

function Live2dAvatar:OnSetText(user_id, show, text, time_sec)
    self:_SetText(user_id, show, text)
    if time_sec > 0 then
        self:_PlayText(user_id, time_sec * 1000)
    end
end

function Live2dAvatar:OnSetBackground(user_id, path)
    if user_id == -1 then
        local tex = hl.ImageTexture_New()
        if tex:InitWithSource(hl.Path(path), true, false) then
            self.bg_sp:GetMaterial():SetTexture(0, tex)
            self.bg_sp:SetWidth(hl.RuntimeWidth())
            local h = hl.RuntimeWidth() * tex:GetHeight() / tex:GetWidth()
            self.bg_sp:SetHeight(h)
            self.bg_ac:GetTransform():SetLocalPosition(hl.Vec3(0, (hl.RuntimeHeight() - h) / 2, 0))
            self.bg_sp:SetVisible(true)
        else
            hl.Class_Discard(tex)
            self.bg_sp:SetVisible(false)
        end
    else
        local model = self.live2d_models[user_id]
        if model then
            local tex = hl.ImageTexture_New()
            if tex:InitWithSource(hl.Path(path), true, false) then
                model.bg_sp:GetMaterial():SetTexture(0, tex)
                model.bg_sp:SetWidth(hl.RuntimeWidth())
                local h = hl.RuntimeWidth() * tex:GetHeight() / tex:GetWidth()
                model.bg_sp:SetHeight(h)
                model.bg_ac:GetTransform():SetLocalPosition(hl.Vec3(0, (hl.RuntimeHeight() - h) / 2, 0))
                if model.visibile == 1 then
                    model.bg_sp:SetVisible(true)
                else
                    model.bg_sp:SetVisible(false)
                end
            else
                hl.Class_Discard(tex)
                model.bg_sp:SetVisible(false)
            end
        end
    end
end

function Live2dAvatar:OnSetMode(user_id, mode)
    local model = self.live2d_models[user_id]
    if model then
        model.mode = mode
    end
end

function Live2dAvatar:OnSetStyle(user_id, style)
    local model = self.live2d_models[user_id]
    if model then
        if model.style ~= style then
            model.style = style
            local trans = model.ac:GetTransform()
            if model.style == 0 then
                local sp_height = model.sp:GetHeight()
                trans:SetLocalPosition(hl.Vec3(0, (sp_height - hl.RuntimeHeight()) *0.5 - hl.RuntimeHeight() * model_y_trans, 0))
                trans:SetLocalScale(hl.Vec3(1))
            elseif model.style == 1 then
                trans:SetLocalPosition(hl.Vec3(0, hl.RuntimeHeight() * 0.12 - hl.RuntimeHeight() * model_y_trans ,0))
                trans:SetLocalScale(hl.Vec3(0.83))
            elseif model.style == 2 then
                trans:SetLocalPosition(hl.Vec3(0))
                trans:SetLocalScale(hl.Vec3(1.3))
            end
        end
    end
end

function Live2dAvatar:OnModelRelease(user_id)
    local model = self.live2d_models[user_id]
    if model then
        model.root_group:RemoveFromWorld(self.node_context.node_manager:GetWorld())
        hl.Class_Release(model.root_group)
        hl.Class_Release(model.sys_text)
        hl.Class_Release(model.text_anim_src)
        self.live2d_models[user_id] = nil
    end
end

function Live2dAvatar:OnSetModelVisibile(user_id, visibile)
    local model = self.live2d_models[user_id]
    if model then
        model.visibile = visibile
        if visibile == 1 then
            model.l2d:SetPause(false)
            model.sp:SetVisible(true)
            model.bg_sp:SetVisible(true)
            model.textbg_sp:SetVisible(true)
            model.text_sp:SetVisible(true)
        else
            model.l2d:SetPause(true)
            model.sp:SetVisible(false)
            model.bg_sp:SetVisible(false)
            model.textbg_sp:SetVisible(false)
            model.text_sp:SetVisible(false)
        end
    end
end

function Live2dAvatar:OnSetMotionStreak(user_id, enabled, config_path)
    self.enable_motion_streak = enabled
    if self.enable_motion_streak then
        local config = cjson.decode(io.load(config_path))
        if config.image then
            local base_path = hl.Path(config_path):ExtractParent()
            local tex = hl.ImageTexture_New()
            if tex:InitWithSource(hl.Path(base_path .. "/" .. config.image)) then
                self.particle:GetMaterial():SetTexture(0, tex)
                if config.image_column then
                    self.uv_affector:SetTileColumn(config.image_column)
                end
                return
            else
                hl.Class_Discard(tex)
            end
        end
    end
end

function Live2dAvatar:OnSetModelSlot(user_id, slot)
    local model = self.live2d_models[user_id]
    if model then
        local ac_trans = model.ac:GetTransform()
        local y = (model.sp:GetHeight() - hl.RuntimeHeight()) * 0.5 - hl.RuntimeHeight() * model_y_trans * model.mac_scale
        local bg_trans = model.bg_ac:GetTransform()
        local bg_y = (hl.RuntimeHeight() - model.bg_sp:GetHeight()) / 2
        local yy = (model.sp:GetHeight() * 0.5 - hl.RuntimeHeight()) * 0.5 - hl.RuntimeHeight() * model_y_trans * model.mac_scale
        if slot == hl.Hl_LIVE2D_SLOT_DEFAULT then
            local sp_height = model.sp:GetHeight()
            ac_trans:SetLocalPosition(hl.Vec3(0, y, 0))
            ac_trans:SetLocalScale(hl.Vec3(1))
            bg_trans:SetLocalPosition(hl.Vec3(0, bg_y, 0))
            bg_trans:SetLocalScale(hl.Vec3(1))
        elseif slot == hl.Hl_LIVE2D_SLOT_LEFT then
            ac_trans:SetLocalPosition(hl.Vec3(-hl.RuntimeWidth() / 4, yy, 0))
            ac_trans:SetLocalScale(hl.Vec3(0.5, 0.5, 1))
            bg_trans:SetLocalPosition(hl.Vec3(-hl.RuntimeWidth() / 4, bg_y - hl.RuntimeHeight() * 0.25, 0))
            bg_trans:SetLocalScale(hl.Vec3(0.5, 0.5, 1))
        elseif slot == hl.Hl_LIVE2D_SLOT_RIGHT then
            ac_trans:SetLocalPosition(hl.Vec3(hl.RuntimeWidth() / 4, yy, 0))
            ac_trans:SetLocalScale(hl.Vec3(0.5, 0.5, 1))
            bg_trans:SetLocalPosition(hl.Vec3(hl.RuntimeWidth() / 4, bg_y - hl.RuntimeHeight() * 0.25, 0))
            bg_trans:SetLocalScale(hl.Vec3(0.5, 0.5, 1))
        elseif slot == Hl_LIVE2D_SLOT_LIVESHOW_1 then

        elseif slot == Hl_LIVE2D_SLOT_LIVESHOW_2 then
        
        elseif slot == Hl_LIVE2D_SLOT_LIVESHOW_3 then

        elseif slot == Hl_LIVE2D_SLOT_LIVESHOW_4 then

        elseif slot == Hl_LIVE2D_SLOT_LIVESHOW_5 then

        elseif slot == Hl_LIVE2D_SLOT_LIVESHOW_6 then

        end
    end
end

function Live2dAvatar:OnSetEnableBreath(user_id, enable)
    local model = self.live2d_models[user_id]
    if model then
        model.l2d:SetEnableBreath(enable)
    end
end

function Live2dAvatar:OnSetEnableEyeBlink(user_id, enable)
    local model = self.live2d_models[user_id]
    if model then
        model.l2d:SetEnableEyeBlink(enable)
    end
end

function Live2dAvatar:OnSetEnableFaceTracker(user_id, enabled)
    local model = self.live2d_models[user_id]
    if model then
        model.enable_face_tracker = enabled
        self:EnableService(Service.SERVICE_FACE_TRACK, enabled)
    end
end

function Live2dAvatar:OnSetEnableAvatarTracker(user_id, enabled)
    self.num_avatar_tracker = self.num_avatar_tracker or 0
    local model = self.live2d_models[user_id]
    if model then
        model.enable_avatar_tracker = model.enable_avatar_tracker or false
        if model.enable_avatar_tracker ~= enabled then
            model.enable_avatar_tracker = enabled
            if enabled then
                self.num_avatar_tracker = self.num_avatar_tracker + 1
                if self.num_avatar_tracker == 1 then
                    self.avatar_tracker:Start()
                end
            else
                self.num_avatar_tracker = self.num_avatar_tracker - 1
                if self.num_avatar_tracker == 0 then
                    self.avatar_tracker:Stop()
                end
            end
        end
    end
end

function Live2dAvatar:OnSetEnableWatermark(user_id, enabled)
    if self.enable_watermark ~= enabled then
        self.enable_watermark = enabled
        if self.enable_watermark then
            local x = (hl.RuntimeWidth() - self.watermark_sp:GetWidth()) / 2
            local y = (hl.RuntimeHeight() - self.watermark_sp:GetHeight()) / 2
            self.watermark_ac:GetTransform():SetLocalPosition(hl.Vec3(x, y, 0))
            self.watermark_sp:SetVisible(true)
        else
            self.watermark_sp:SetVisible(false)
        end
    end
end

local MouthLeft = 90
local MouthRight = 84
local MouthTop = 98
local MouthBottom = 102
local REyeTop = 72
local REyeBottom = 73
local REyeLeft = 52
local REyeRight = 55
local REyeCenter = 74
local LEyeTop = 75
local LEyeBottom = 76
local LEyeLeft = 58
local LEyeRight = 61
local LEyeCenter = 77
local NoseTop = 43
local NoseBottom = 46
local FaceLeft = 32
local FaceRight = 0
local LEyebrowTop = 40
local LEyebrowLeft = 38
local LEyebrowRight = 42
local REyebrowTop = 35
local REyebrowLeft = 33
local REyebrowRight = 37

local function ParamMapping(dst_min, dst_max, src_min, src_max, src)
    src = hl.Clamp(src, src_min, src_max)
    local d = (src - src_min) / (src_max - src_min)
    local dst = dst_min + (dst_max - dst_min) * d
    return dst
end
function Live2dAvatar:_UpdateFaceTracker()
    for user_id, model in pairs(self.live2d_models) do
        if model.enable_face_tracker then
            local face_infos = self.node_context.node_manager:GetService(Service.SERVICE_FACE_TRACK):GetFaceInfos()
            if not face_infos then
                LOGW("No face info:", user_id)
                return
            end
            
            if face_infos.face_num > 0 then
                local fi = face_infos.faces[face_infos.valid_face_ids[1]].original_face_data
                local nose_top = fi.key_points[NoseTop]
                local face_width = fi.key_points[FaceRight]:Distance(fi.key_points[FaceLeft])
                local nose_height = nose_top:Distance(fi.key_points[NoseBottom])
                local mouth_width = fi.key_points[MouthLeft]:Distance(fi.key_points[MouthRight])
                local mouth_height = fi.key_points[MouthTop]:Distance(fi.key_points[MouthBottom])
                local leye_left = fi.key_points[LEyeLeft]
                local leye_right = fi.key_points[LEyeRight]
                local leye_top = fi.key_points[LEyeTop]
                local leye_bottom = fi.key_points[LEyeBottom]
                local leye_center = fi.key_points[LEyeCenter]
                local reye_left = fi.key_points[REyeLeft]
                local reye_right = fi.key_points[REyeRight]
                local reye_top = fi.key_points[REyeTop]
                local reye_bottom = fi.key_points[REyeBottom]
                local reye_center = fi.key_points[REyeCenter]

                local leyebrow_top = fi.key_points[LEyebrowTop]
                local leyebrow_left = fi.key_points[LEyebrowLeft]
                local leyebrow_right = fi.key_points[LEyebrowRight]
                local reyebrow_top = fi.key_points[REyebrowTop]
                local reyebrow_left = fi.key_points[REyebrowLeft]
                local reyebrow_right = fi.key_points[REyebrowRight]

                local lbrow_y = (nose_top.y - leyebrow_left.y) / nose_height
                local rbrow_y = (nose_top.y - reyebrow_right.y) / nose_height

                local lbrow_x = (leyebrow_left.x - nose_top.x) / face_width
                local rbrow_x = (nose_top.x - reyebrow_right.x) / face_width

                -- local lbrow_rot = leyebrow_right.y - leyebrow_left.y
                -- local rbrow_rot = reyebrow_left.y - reyebrow_right.y
                local angle = fi.rot2d:EulerAngle()
                local tween = 0.382
                if not model.lbrow_y then
                    model.lbrow_y = ParamMapping(-1, 1, 0.4, 0.65, lbrow_y)
                else
                    lbrow_y = ParamMapping(-1, 1, 0.4, 0.65, lbrow_y)
                    model.lbrow_y = hl.Lerp(model.lbrow_y, lbrow_y, tween)
                end
                if not model.rbrow_y then
                    model.rbrow_y = ParamMapping(-1, 1, 0.4, 0.65, rbrow_y)
                else
                    rbrow_y = ParamMapping(-1, 1, 0.4, 0.65, rbrow_y)
                    model.rbrow_y = hl.Lerp(model.rbrow_y, rbrow_y, tween)
                end
                if not model.angle_y then
                    model.angle_y = ParamMapping(-30, 30, -0.8, 0.8, angle.y)
                else
                    local angle_y = ParamMapping(-30, 30, -0.8, 0.8, angle.y)
                    model.angle_y = hl.Lerp(model.angle_y, angle_y, tween)
                end
                if not model.angle_x then
                    model.angle_x = ParamMapping(-30, 30, -0.3, 0.3, -angle.x)
                else
                    local angle_x = ParamMapping(-30, 30, -0.3, 0.3, -angle.x)
                    model.angle_x = hl.Lerp(model.angle_x, angle_x, tween)
                end
                if not model.angle_z then
                    model.angle_z = ParamMapping(-30, 30, -0.3, 0.3, -angle.z)
                else
                    local angle_z = ParamMapping(-30, 30, -0.3, 0.3, -angle.z)
                    model.angle_z = hl.Lerp(model.angle_z, angle_z, tween)
                end
                if not self.leye_open then
                    model.leye_open = ParamMapping(0, 2, 0.13, 0.25, (leye_bottom.y - leye_top.y) / nose_height)
                else
                    local leye_open = ParamMapping(0, 2, 0.13, 0.25, (leye_bottom.y - leye_top.y) / nose_height)
                    model.leye_open = hl.Lerp(model.leye_open, leye_open, tween)
                end
                if not self.reye_open then
                    model.reye_open = ParamMapping(0, 2, 0.13, 0.25, (reye_bottom.y - reye_top.y) / nose_height)
                else
                    local reye_open = ParamMapping(0, 2, 0.13, 0.25, (reye_bottom.y - reye_top.y) / nose_height)
                    model.reye_open = hl.Lerp(model.reye_open, reye_open, tween)
                end
                if not self.mouth_form then
                    model.mouth_form = ParamMapping(-1, 1, 0.30, 0.45, mouth_width / face_width)
                else
                    local mouth_form = ParamMapping(-1, 1, 0.30, 0.45, mouth_width / face_width)
                    model.mouth_form = hl.Lerp(model.mouth_form, mouth_form, tween)
                end
                if not self.mouth_open_y then
                    model.mouth_open_y = ParamMapping(0, 1, 0.15, 0.8, mouth_height / nose_height)
                else
                    local mouth_open_y = ParamMapping(0, 1, 0.15, 0.8, mouth_height / nose_height)
                    model.mouth_open_y = hl.Lerp(model.mouth_open_y, mouth_open_y, tween)
                end
                if not self.eye_ball_x then
                    model.eye_ball_x = ParamMapping(1, -1, 0.45, 0.55, (leye_center.x - leye_left.x) / (leye_right.x - leye_left.x))
                else
                    local eye_ball_x = ParamMapping(1, -1, 0.45, 0.55, (leye_center.x - leye_left.x) / (leye_right.x - leye_left.x))
                    model.eye_ball_x = hl.Lerp(model.eye_ball_x, eye_ball_x, tween)
                end
                if not self.eye_ball_y then
                    model.eye_ball_y = ParamMapping(-1, 1, 0.45, 0.55, (leye_center.y - leye_bottom.y) / (leye_top.y - leye_bottom.y))
                else
                    local eye_ball_y = ParamMapping(-1, 1, 0.45, 0.55, (leye_center.y - leye_bottom.y) / (leye_top.y - leye_bottom.y))
                    model.eye_ball_y = hl.Lerp(model.eye_ball_y, eye_ball_y, tween)
                end

                self:OnSetParameter(user_id, "ParamBrowLY", model.lbrow_y)
                self:OnSetParameter(user_id, "ParamBrowRY", model.rbrow_y)
                -- self:OnSetParameter(1, "ParamBrowLX", ParamMapping(-1, 1, 0.08, 0.12, lbrow_x))
                -- self:OnSetParameter(1, "ParamBrowRX", ParamMapping(-1, 1, 120, 150, rbrow_x))
                -- self:OnSetParameter(1, "ParamBrowLAngle", ParamMapping(-1, 1, 120, 150, lbrow_rot))
                -- self:OnSetParameter(1, "ParamBrowRAngle", ParamMapping(-1, 1, 120, 150, rbrow_rot))
                self:OnSetParameter(user_id, "ParamAngleX", model.angle_y)
                self:OnSetParameter(user_id, "ParamAngleY", model.angle_x)
                self:OnSetParameter(user_id, "ParamAngleZ", model.angle_z)
                self:OnSetParameter(user_id, "ParamEyeLOpen", model.leye_open)
                self:OnSetParameter(user_id, "ParamEyeROpen", model.reye_open)
                self:OnSetParameter(user_id, "ParamMouthForm", model.mouth_form)
                self:OnSetParameter(user_id, "ParamMouthOpenY", model.mouth_open_y)
                self:OnSetParameter(user_id, "ParamEyeBallX", model.eye_ball_x)
                self:OnSetParameter(user_id, "ParamEyeBallY", model.eye_ball_y)
            end
        end
    end
end

local tween = 1 - 0.382 / 2
function Live2dAvatar:_SetAvatarParameter(user_id, name, dst_min, dst_max, src_min, src_max, src, i_tween)
    local mp = self.live2d_models[user_id].parameters
    if not mp[name] then
        mp[name] = ParamMapping(dst_min, dst_max, src_min, src_max, src)
    else
        local value = ParamMapping(dst_min, dst_max, src_min, src_max, src)
        mp[name] = hl.Lerp(mp[name], value, i_tween or tween)
    end
    self:OnSetTempParameter(user_id, name, mp[name])
end

function Live2dAvatar:_UpdateAvatarTracker()
    if self.num_avatar_tracker and self.num_avatar_tracker > 0 then
        self.avatar_tracker:Update()
        if self.avatar_tracker:GetAvatarCount() == 0 then
            return
        end
        local avatar_expressions = {}
        for i = 0, 45 do
            table.insert(avatar_expressions, self.avatar_tracker:GetExpressionValue(0, i))
        end
        local angle = self.avatar_tracker:GetRotationValue(0):EulerAngle()
        if angle.z > 0 then
            angle.z = hl.kPi - angle.z
        elseif angle.z < 0 then
            angle.z = -(angle.z + hl.kPi)
        end
        
        local leye_open = 1 - avatar_expressions[1]
        local reye_open = 1 - avatar_expressions[2]
        local leye_squint = avatar_expressions[3]
        local reye_squint = avatar_expressions[4]
        local leye_down = avatar_expressions[5]
        local reye_down = avatar_expressions[6]
        local leye_inside = avatar_expressions[7]
        local reye_inside = avatar_expressions[8]
        local eye_open = (avatar_expressions[9] + avatar_expressions[10]) / 2 * 2.5
        local leye_outside = avatar_expressions[11]
        local reye_outside = avatar_expressions[12]
        local leye_up = avatar_expressions[13]
        local reye_up = avatar_expressions[14]

        local eye_ball_up = (leye_up + reye_up) / 2
        local eye_ball_down = (leye_down + reye_down) / 2 
        local eye_ball_left = (leye_outside + reye_inside) / 2
        local eye_ball_right = (leye_inside + reye_outside) / 2

        local lbrow_down = avatar_expressions[15]
        local brow_center_up = avatar_expressions[17] * 2
        local rbrow_down = avatar_expressions[16]
        local lbrow_up = avatar_expressions[18]
        local rbrow_up = avatar_expressions[19]
        local mouth_open = avatar_expressions[22]
        local lmouth_down = avatar_expressions[26]
        local rmouth_down = avatar_expressions[27]
        local lmouth_up = avatar_expressions[28]
        local rmouth_up = avatar_expressions[29]
        local lips_pucker = avatar_expressions[40]
        local sneer = avatar_expressions[43] * 2
        
        local brow_tween = 1 - 0.382
        local eye_ball_tween = 0.382 / 2
        for user_id, model in pairs(self.live2d_models) do
            if model.enable_avatar_tracker then
                self:_SetAvatarParameter(user_id, "ParamEyeForm", 0, 1, 0, 1, eye_open)
                
                if hl.IsPlatformAndroid() then
                    self:_SetAvatarParameter(user_id, "ParamEyeLOpen", 0, 1, 0, 1, leye_open)
                    self:_SetAvatarParameter(user_id, "ParamEyeROpen", 0, 1, 0, 1, reye_open)
                    self:_SetAvatarParameter(user_id, "ParamEyeLSmile", 0, 1, 0, 1, leye_squint)
                    self:_SetAvatarParameter(user_id, "ParamEyeRSmile", 0, 1, 0, 1, reye_squint)
                    self:_SetAvatarParameter(user_id, "ParamAngleX", -30, 30, -0.5, 0.5, angle.y)
                    self:_SetAvatarParameter(user_id, "ParamAngleZ", -30, 30, -0.5, 0.5, -angle.x)
                    if rbrow_down > 0 then
                        self:_SetAvatarParameter(user_id, "ParamBrowRY2", 0, -1, 0, 1, rbrow_down, brow_tween)
                    elseif rbrow_up > 0 then
                        self:_SetAvatarParameter(user_id, "ParamBrowRY2", 0, 1, 0, 1, rbrow_up, brow_tween)
                    end
                    if lbrow_down > 0 then
                        self:_SetAvatarParameter(user_id, "ParamBrowLY2", 0, -1, 0, 1, lbrow_down, brow_tween)
                    elseif lbrow_up > 0 then
                        self:_SetAvatarParameter(user_id, "ParamBrowLY2", 0, 1, 0, 1, lbrow_up, brow_tween)
                    end
                    if lmouth_down > 0 then
                        self:OnSetTempParameter(user_id, "ParamMouthL", 0, -1, 0, 1, lmouth_down)
                    elseif lmouth_up > 0 then
                        self:OnSetTempParameter(user_id, "ParamMouthL", 0, 1, 0, 1, lmouth_up)
                    end
                    if rmouth_down > 0 then
                        self:_SetAvatarParameter(user_id, "ParamMouthR", 0, -1, 0, 1, rmouth_down)
                    elseif rmouth_up > 0 then
                        self:_SetAvatarParameter(user_id, "ParamMouthR", 0, 1, 0, 1, rmouth_up)
                    end
                else
                    self:_SetAvatarParameter(user_id, "ParamEyeLOpen", 0, 1, 0, 1, reye_open)
                    self:_SetAvatarParameter(user_id, "ParamEyeROpen", 0, 1, 0, 1, leye_open)
                    self:_SetAvatarParameter(user_id, "ParamEyeLSmile", 0, 1, 0, 1, reye_squint)
                    self:_SetAvatarParameter(user_id, "ParamEyeRSmile", 0, 1, 0, 1, leye_squint)
                    self:_SetAvatarParameter(user_id, "ParamAngleX", -30, 30, -0.5, 0.5, -angle.y)
                    self:_SetAvatarParameter(user_id, "ParamAngleZ", -30, 30, -0.5, 0.5, angle.x)
                    if lbrow_down > 0 then
                        self:_SetAvatarParameter(user_id, "ParamBrowRY2", 0, -1, 0, 1, lbrow_down, brow_tween)
                    elseif lbrow_up > 0 then
                        self:_SetAvatarParameter(user_id, "ParamBrowRY2", 0, 1, 0, 1, lbrow_up, brow_tween)
                    end
                    if rbrow_down > 0 then
                        self:_SetAvatarParameter(user_id, "ParamBrowLY2", 0, -1, 0, 1, rbrow_down, brow_tween)
                    elseif rbrow_up > 0 then
                        self:_SetAvatarParameter(user_id, "ParamBrowLY2", 0, 1, 0, 1, rbrow_up, brow_tween)
                    end
                    if rmouth_down > 0 then
                        self:OnSetTempParameter(user_id, "ParamMouthL", 0, -1, 0, 1, rmouth_down)
                    elseif rmouth_up > 0 then
                        self:OnSetTempParameter(user_id, "ParamMouthL", 0, 1, 0, 1, rmouth_up)
                    end
                    if lmouth_down > 0 then
                        self:_SetAvatarParameter(user_id, "ParamMouthR", 0, -1, 0, 1, lmouth_down)
                    elseif lmouth_up > 0 then
                        self:_SetAvatarParameter(user_id, "ParamMouthR", 0, 1, 0, 1, lmouth_up)
                    end
                end
                self:_SetAvatarParameter(user_id, "ParamAngleY", -30, 30, -0.5, 0.5, angle.z)
                self:_SetAvatarParameter(user_id, "ParamMouthForm", 1, -1, 0, 1, lips_pucker)
                self:_SetAvatarParameter(user_id, "ParamMouthOpenY", 0, 2, 0, 1, mouth_open)
                if eye_ball_left > 0 then
                    self:_SetAvatarParameter(user_id, "ParamEyeBallX", 0, 1, 0, 1, eye_ball_left, eye_ball_tween)
                elseif eye_ball_right > 0 then
                    self:_SetAvatarParameter(user_id, "ParamEyeBallX", 0, -1, 0, 1, eye_ball_right, eye_ball_tween)
                end
                if eye_ball_up > 0 then
                    self:_SetAvatarParameter(user_id, "ParamEyeBallY", 0, 1, 0, 1, eye_ball_up, eye_ball_tween)
                elseif eye_ball_down > 0 then
                    self:_SetAvatarParameter(user_id, "ParamEyeBallY", 0, -1, 0, 1, eye_ball_down, eye_ball_tween)
                end
                if brow_center_up > 0 then
                    self:_SetAvatarParameter(user_id, "ParamBrowLForm", 0, 1, 0, 1, brow_center_up, brow_tween)
                    self:_SetAvatarParameter(user_id, "ParamBrowRForm", 0, 1, 0, 1, brow_center_up, brow_tween)
                elseif sneer > 0 then
                    self:_SetAvatarParameter(user_id, "ParamBrowLForm", 0, -1, 0, 1, sneer, brow_tween)
                    self:_SetAvatarParameter(user_id, "ParamBrowRForm", 0, -1, 0, 1, sneer, brow_tween)
                end
            end
        end
    end
end

function Live2dAvatar:InitTestModel(user_id, avatar_file, user_data_file)
    local json_text = io.load((self.work_dir / hl.Path(user_data_file .. "/avatar_user_data.json")):CStr())
    local cfg = cjson.decode(json_text)
    for _, v in ipairs(cfg.motions) do
        v.motionId = (self.work_dir / (avatar_file .. "/" .. cfg.avatarId .. "/motions/" .. v.motionId .. "/" .. v.motionId .. ".motion3.json")):CStr()
    end
    for _, v in ipairs(cfg.expressions) do
        v.expressionId = (self.work_dir / (avatar_file .. "/" .. cfg.avatarId .. "/expressions/" .. v.expressionId .. "/" .. v.expressionId .. ".exp3.json")):CStr()
    end
    for _, v in ipairs(cfg.components) do
        v.componentId = (self.work_dir / (avatar_file .. "/" .. cfg.avatarId .. "/components/" .. v.type .. "/" .. v.componentId .. "/" .. v.componentId .. ".png")):CStr()
    end
    for _, action in ipairs(cfg.actions) do
        for k, trigger in ipairs(action.triggers) do
            if trigger.type == "motions" then
                for i, motion in ipairs(trigger.value) do
                    trigger.value[i] = (self.work_dir / (avatar_file .. "/" .. cfg.avatarId .. "/motions/" .. motion .. "/" .. motion .. ".motion3.json")):CStr()
                end
            elseif trigger.type == "expressions" then
                for i, expression in ipairs(trigger.value) do
                    trigger.value[i] = (self.work_dir / (avatar_file .. "/" .. cfg.avatarId .. "/expressions/" .. expression .. "/" .. expression .. ".exp3.json")):CStr()
                end
            elseif trigger.type == "components" then
                for i, component in ipairs(trigger.replace) do
                    trigger.replace[i] = (self.work_dir / (avatar_file .. "/" .. cfg.avatarId .. "/components/" .. trigger.value[i] .. "/" .. component .. "/" .. component .. ".png")):CStr()
                end
            end
        end
    end
    cfg.physicsId = (self.work_dir / (avatar_file .. "/" .. cfg.avatarId .. "/" .. cfg.avatarId .. ".physics3.json")):CStr()
    cfg.avatarId = (self.work_dir / (avatar_file .. "/" .. cfg.avatarId .. "/" .. cfg.avatarId .. ".moc3")):CStr()
    self:OnModelInitWithJson(user_id, cfg)
    -- self:_SetText(user_id, 1, "哈哈哈哈哈")
    -- self:OnSetBackground(user_id, (self.work_dir / "Textures/Grid.png"):CStr())
    -- self:_PlayText(user_id, 1500)
    -- self:OnSetStyle(user_id, 2)
end

function Live2dAvatar:InitTestModelNew()
    local json_text = io.load((self.work_dir / hl.Path("Live2DNew/Test/avatar_user_data.json")):CStr())
    local cfg = cjson.decode(json_text)
    for _, v in ipairs(cfg.motions) do
        v.motionId = (self.work_dir / ("Live2DNew/" .. cfg.avatarId .. "/motions/" .. v.motionId .. "/" .. v.motionId .. ".motion3.json")):CStr()
    end
    for _, v in ipairs(cfg.components) do
        v.componentId = (self.work_dir / ("Live2DNew/" .. cfg.avatarId .. "/components/" .. v.componentId)):CStr()
    end
    for _, action in ipairs(cfg.actions) do
        for k, trigger in ipairs(action.triggers) do
            if trigger.type == "motions" then
                for i, motion in ipairs(trigger.value) do
                    trigger.value[i] = (self.work_dir / ("Live2DNew/" .. cfg.avatarId .. "/motions/" .. motion .. "/" .. motion .. ".motion3.json")):CStr()
                end
            elseif trigger.type == "components" then
                for i, component in ipairs(trigger.replace) do
                    trigger.replace[i] = (self.work_dir / ("Live2DNew/" .. cfg.avatarId .. "/components/" .. trigger.value[i] .. "/" .. component .. "/" .. component .. ".png")):CStr()
                end
            end
        end
    end
    cfg.avatarId = (self.work_dir / ("Live2DNew/" .. cfg.avatarId .. "/" .. cfg.avatarId .. ".moc3")):CStr()
    self:OnModelInitWithJson(100, cfg)
    self:_SetText(100, 1, "哈哈哈哈哈")
    self:OnSetBackground(100, (self.work_dir / "Textures/Grid.png"):CStr())
    self:_PlayText(100, 1500)
end

return Live2dAvatar
