--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local Service = import "Service"

local MultiOutputService = define ("MultiOutputService", Service)

LOGD("MultiOutputService has been defined ..., parent is", Service)

function MultiOutputService:_OnInitialize()
    self.service_type = Service.SERVICE_MULTI_OUTPUT

    self.output_texture_copier = hl.TextureCopy_New()
    hl.Class_Retain(self.output_texture_copier)
    self.output_texture_copier_inited = false
    self.need_render = true

    self.frame_view_id_list = hl.U16Array()
end

function MultiOutputService:SetNeedRender(need)
    self.need_render = need
end

function MultiOutputService:_Prepare()
    if self.output_texture_copier_inited == true then
        return
    end

    self.output_texture_copier:InitWithSetting( self.attach_node.node_context.node_manager:GetFixedTexture(
            self.attach_node.node_context.node_manager.INPUT_TEXTURE_0, true), self.attach_node.node_context.node_manager:GetFixedTexture(
            self.attach_node.node_context.node_manager.OUTPUT_TEXTURE_1, false))

    self.output_texture_copier_inited = true
end

function MultiOutputService:_OnUpdate(dt)
    self:_Prepare()

    if self.need_render then
        if self.output_texture_copier:PrepareRender() then
            local count = self.output_texture_copier:GetRenderViewCount()
            self.frame_view_id_list:clear()
            for k = 1, count do
                self.frame_view_id_list:push_back(hl.GPRenderDrive_AllocFrameView())
            end
            self.output_texture_copier:Render(self.frame_view_id_list)
        end
    end
end

function MultiOutputService:_OnStart()
    local show_bg_service = self.attach_node:GetService(Service.SERVICE_SHOW_BG)
    show_bg_service:SetBGTexture(self.attach_node.node_context.node_manager:GetFixedTexture(
            self.attach_node.node_context.node_manager.OUTPUT_TEXTURE_1, false))
end

function MultiOutputService:_OnStop()
    local show_bg_service = self.attach_node:GetService(Service.SERVICE_SHOW_BG)
    show_bg_service:SetBGTexture(self.attach_node.node_context.node_manager:GetFixedTexture(
            self.attach_node.node_context.node_manager.INPUT_TEXTURE_0, true))
end

function MultiOutputService:_OnShutdown()
    hl.Class_Release(self.output_texture_copier)
    self.output_texture_copier = nil
end

return MultiOutputService