--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local NodeLayer = import "NodeLayer"

local NormalNodeLayer = define ("NormalNodeLayer", NodeLayer)

LOGD("NormalNodeLayer has been defined, parent is ", NodeLayer)

function NormalNodeLayer:_OnInitialize()
end

function NormalNodeLayer:_OnAddNode(node_type, node_id)

    local new_node = self:_RequireNode(node_type, node_id)
    if new_node == nil then
        CHECK(false, "NormalNodeLayer: _OnAddNode: _RequireNode:", node_type, "fail !!!")
        return nil, nil
    end

    local new_node_pos = -1
    local new_node_inserted = false
    for pos, node in ipairs(self.ordered_node_list) do
        if node:GetNodeOrder() >= new_node:GetNodeOrder() then
            table.insert(self.ordered_node_list, pos, new_node)
            new_node_inserted = true
            new_node_pos = pos
            break
        end
    end

    if not new_node_inserted then
        table.insert(self.ordered_node_list, new_node)
        new_node_pos = #self.ordered_node_list
    end

    self.node_manager:AddNodeImpl(new_node, new_node_pos, self)

    LOGD("NormalNodeLayer: Add Node:", node_type, node_id, "pos:", new_node_pos)

    self:UpdateNodesPos()

    return true, new_node
end

function NormalNodeLayer:_OnRemoveNode(node_type, node_id)
    local node_pos = self.node_manager:RemoveNodeImpl(node_type, node_id)

    if nil == node_pos then
        return false
    end

    CHECK(node_pos > 0, "NormalNodeLayer: _OnRemoveNode, Node's", node_type, node_id, "invalid pos:", node_pos, "!!!")

    LOGD("NormalNodeLayer: _OnRemoveNode, Node's", node_type, node_id, "pos:", node_pos)
    table.remove(self.ordered_node_list, node_pos)

    self:UpdateNodesPos()

    return true
end

return NormalNodeLayer