--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local Service = import "Service"

local OccluderService = define ("OccluderService", Service)

LOGD("OccluderService has been defined ..., parent is", Service)

function OccluderService:_OnInitialize()
    self.service_type = Service.SERVICE_OCCLUDER

    self.kHuman_Head_Layer = -20000
    self.human_head_pose_info = { num = 0, head_transforms = {} }

    self.enable_occluder = true
end

function OccluderService:EnableOccluder(enable)
    self.enable_occluder = enable
end

function OccluderService:_fillHumanHead(num)
    if num > Service.MAX_FACE_NUM then
        CHECK(false, "OccluderService: _fillHumanHead: num:", num, "> Service.MAX_FACE_NUM", Service.MAX_FACE_NUM, "!!!")
        return
    end

    if #self.human_heads >= num then
        return
    end

    LOGD("OccluderService: _fillHumanHead num:", #self.human_heads, "->", num)

    local begin_index = #self.human_heads + 1
    local end_index = num

    local new_human_head = nil
    for i = begin_index, end_index do
        new_human_head = {}
        new_human_head.actor = self.attach_node:AllocActor(hl.Named("HumanHead" .. i))
        new_human_head.mesh = hl.Mesh_New()
        new_human_head.mesh:SetMeshData(self.human_head_mesh_data)
        new_human_head.mesh:SetLayer(self.kHuman_Head_Layer)
        local mtl = hl.Mesh_CreateSimpleMaterial()
        local render_state = mtl:GetRenderState()
        render_state.write = hl.GPWriteFlag_kZ
        mtl:SetRenderState(render_state)
        mtl:SetBaseColorTex(self.attach_node:GetDefaultTexture())
        new_human_head.mesh:SetMaterial(mtl)
        new_human_head.mesh:SetVisible(false)
        new_human_head.attach_actor = self.attach_node:AllocActor(hl.Named("HumanHead" .. i .. "_attach"))
        new_human_head.actor:AddComponent(new_human_head.mesh)
        self.attach_node:AddToWorld(new_human_head.actor)
        self.attach_node:AddToWorld(new_human_head.attach_actor)
        new_human_head.actor:GetTransform():SetParent(new_human_head.attach_actor:GetTransform())
        self.attach_node:SetRenderLayer(new_human_head.mesh, self.attach_node.LAYER_3D_0)
        table.insert(self.human_heads, new_human_head)
    end
end

function OccluderService:_OnStop()
    self:EnableService(Service.SERVICE_FACE_TRACK, false)
    local face_track_service = self:GetService(Service.SERVICE_FACE_TRACK)
    face_track_service:EnablePose3D(false)
    self:_SetHumanHeadsVisibility(false)
end

function OccluderService:_SetHumanHeadsVisibility(visible)
    for _, human_head in ipairs(self.human_heads) do
        human_head.mesh:SetVisible(visible)
    end
end

function OccluderService:_InitHumanHead()
    LOGD("OccluderService: _InitHumanHead ...")

    self.human_head_mesh_data = hl.MeshData_CreateShape(hl.kSHAPE_HUMAN_HEAD)

    self.human_head_inited = true
    self.human_heads = {}
end

function OccluderService:_OnStart()
    self:EnableService(Service.SERVICE_FACE_TRACK, true)

    local face_track_service = self:GetService(Service.SERVICE_FACE_TRACK)
    face_track_service:EnablePose3D(true)

    if not self.human_head_inited then
        self:_InitHumanHead()
    end
end

function OccluderService:GetHumanHeadPoseInfo()
    return self.human_head_pose_info
end

function OccluderService:_OnUpdate(dt)

    local face_num = 0

    local face_infos = self.attach_node:GetService(Service.SERVICE_FACE_TRACK):GetFaceInfos()

    if face_infos ~= nil then
        face_num = face_infos.face_num
    end

    self.human_head_pose_info.num = face_num

    local face_info = nil

    self:_fillHumanHead(face_num)

    self:_SetHumanHeadsVisibility(false)

    local _, camera_actor = self.attach_node.node_context.node_manager:GetDefaultCamera3d();

    local index = 1
    local head = nil

    local camera_position = hl.Vec3()
    local camera_rotation = hl.Quat()
    local camera_scale = hl.Vec3()

    camera_actor:GetTransform():GetLocalMatrix():Decompose(camera_position, camera_rotation, camera_scale)

    for _, face_id in ipairs(face_infos.valid_face_ids) do
        face_info = face_infos.faces[face_id]

        head = self.human_heads[index]
        if self.enable_occluder then
            head.mesh:SetVisible(true)
        end

        head.attach_actor:GetTransform():SetLocalPosition(camera_position)
        head.attach_actor:GetTransform():SetLocalRotation(camera_rotation)
        head.attach_actor:GetTransform():SetLocalScale(camera_scale)

        head.actor:GetTransform():SetLocalPosition(face_info.original_face_data.pos3d)
        head.actor:GetTransform():SetLocalRotation(face_info.original_face_data.rot3d)
        head.actor:GetTransform():SetLocalScale(face_info.original_face_data.scale3d)

        self.human_head_pose_info.head_transforms[face_id] = {rotation = head.actor:GetTransform():GetGlobalRotation(),
                                                              scale = head.actor:GetTransform():GetGlobalScale(),
                                                              position = head.actor:GetTransform():GetGlobalPosition()}

        index = index + 1
    end
end

return OccluderService