--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local Node = import "Node"
local Service = import "Service"
local ServiceManager = import "ServiceManager"

local RootNode = define ("RootNode", Node)

LOGD("RootNode has been defined ..., parent is", Node)

function RootNode:AttachServiceManager(service_manger)
    self.service_manger = service_manger
end

function RootNode:_OnInitialize()
    LOGD("RootNode has been initialized ...")
    self.service_manger:Initialize(self)

    self:EnableService(Service.SERVICE_SHOW_BG, true)

    return true
end

function RootNode:_OnSwitchState(new_state)
    self.service_manger:SwitchState(new_state)
end

function RootNode:_OnShutdown()
    self:EnableService(Service.SERVICE_SHOW_BG, false)

    self.service_manger:Shutdown()
end

function RootNode:_OnNodeEvent(msg, message_source)
    local action = msg:PopInt32()

    local event_consumed = self.service_manger:OnEvent(action, msg, message_source)

    if event_consumed then
        return true
    end

    if action == hl.Hl_MSG_SCRIPT_NODE_LOGIC_ENABLE_SERVICE then
        local service_type = msg:PopInt32()
        local enable = msg:PopInt8()

        if enable == 1 then
            self:EnableService(service_type, true)
            LOGD("RootNode: _OnNodeEvent: Hl_MSG_SCRIPT_NODE_LOGIC_ENABLE_SERVICE: enable service: type:", service_type)
        else
            self:EnableService(service_type, false)
            LOGD("RootNode: _OnNodeEvent: Hl_MSG_SCRIPT_NODE_LOGIC_ENABLE_SERVICE: disable service: type:", service_type)
        end
    end

    return true
end

function RootNode:BeginFrame(dt)
    self.service_manger:BeginFrame(dt)
end

function RootNode:EndFrame(dt)
    self.service_manger:EndFrame(dt)
end

function RootNode:_OnUpdate(dt)
    self.service_manger:Update(dt)
end

return RootNode