local Service = import "Service"

local ShowBGService = define ("ShowBGService", Service)

LOGD("ShowBGService has been defined ..., parent is", Service)

ShowBGService.kFitFill = 0
ShowBGService.kFitCustom = 1

function ShowBGService:_OnInitialize()
    self.service_type = Service.SERVICE_SHOW_BG

    self.bg_containers = {}

    -- TODO: android默认关闭，为了避免某些没有纹理输入的case, 在某些手机上直接导致花屏
    if hl.IsPlatformiOS() then
        self:EnableBG(self.node_manager.INPUT_TEXTURE_0, true)
    end
end

function ShowBGService:_OnRuntimeReset()
    local runtime_width = hl.RuntimeWidth(hl.kSCREEN_EQUAL)
    local runtime_height = hl.RuntimeHeight(hl.kSCREEN_EQUAL)

    LOGD("ShowBGService: _OnRuntimeReset: new runtime width/height:", runtime_width, runtime_height)

    for _, bg_container in pairs(self.bg_containers) do
        if bg_container.fit_mode == ShowBGService.kFitFill then
            bg_container.bg_sprite:SetWidth(runtime_width)
            bg_container.bg_sprite:SetHeight(runtime_height)
        end
    end

end

function ShowBGService:GetBG(fixed_texture_slot)
    CHECK(fixed_texture_slot ~= nil, "ShowBGService: GetBG: invalid fixed texture slot !!!")
    local bg_container = self.bg_containers[fixed_texture_slot]
    if (bg_container == nil) then
        CHECK(bg_container ~= nil, "ShowBGService: GetBG: fixed_texture_slot:",
                fixed_texture_slot , "is not created yet !!!")
        return nil
    end

    return bg_container
end

function ShowBGService:ShowBG(fixed_texture_slot, show)
    CHECK(fixed_texture_slot ~= nil, "ShowBGService: ShowBG: invalid fixed texture slot !!!")

    if show == nil then
        show = true
    end

    local bg_container = self.bg_containers[fixed_texture_slot]
    if (bg_container == nil) then
        return
    end

    local bg = self.bg_containers[fixed_texture_slot]

    bg.enable = show
    bg.bg_sprite:SetVisible(show)
end

function ShowBGService:EnableBG(fixed_texture_slot, enable, fit_mode)
    CHECK(fixed_texture_slot ~= nil, "ShowBGService: EnableBG: invalid fixed texture slot !!!")

    if enable == nil then
        enable = true
    end

    if fit_mode == nil then
        fit_mode = ShowBGService.kFitFill
    end

    local bg_container = self.bg_containers[fixed_texture_slot]
    if (bg_container ~= nil) then
        bg_container.enable = enable
        bg_container.fit_mode = fit_mode
        return
    end

    local new_bg_container = {}

    new_bg_container.fit_mode = fit_mode
    new_bg_container.bg_actor, new_bg_container.bg_sprite = self.attach_node:AllocSpriteActor(hl.Named("BG" .. fixed_texture_slot))

    new_bg_container.fixed_texture = self.node_manager:GetFixedTexture(
            fixed_texture_slot, true)
    new_bg_container.bg_texture = new_bg_container.fixed_texture
    new_bg_container.bg_sprite:GetMaterial():SetBaseColorTex(new_bg_container.fixed_texture)

    local runtime_width = hl.RuntimeWidth(hl.kSCREEN_EQUAL)
    local runtime_height = hl.RuntimeHeight(hl.kSCREEN_EQUAL)

    new_bg_container.bg_sprite:SetWidth(runtime_width)
    new_bg_container.bg_sprite:SetHeight(runtime_height)
    --FIXME: RuntimeWidth RuntimeHeight may change
    if runtime_height == 0 or runtime_width == 0 then
        LOGW("ShowBGService: EnableBG: fixed_texture_slot:", fixed_texture_slot,
                "width/height is not valid !!!", runtime_width, runtime_height)
    end

    new_bg_container.bg_sprite:SetVisible(enable)
    new_bg_container.bg_sprite:SetLayer(-10000)
    new_bg_container.enable = enable
    self.attach_node:SetRenderLayer(new_bg_container.bg_sprite, self.attach_node.LAYER_BG_2D)
    self.attach_node:AddToWorld(new_bg_container.bg_actor)

    self.bg_containers[fixed_texture_slot] = new_bg_container

    LOGD("ShowBGService: enableBG: fixed texture slot:", fixed_texture_slot,
            "width:", new_bg_container.fixed_texture_width,
            "height:", new_bg_container.fixed_texture_height,
            "enable:", enable, "fixed id", new_bg_container.fixed_texture:GetID())
end

function ShowBGService:SetBGTexture(texture, fixed_texture_slot)
    if fixed_texture_slot == nil then
        fixed_texture_slot = self.node_manager.INPUT_TEXTURE_0
    end

    local bg_container = self.bg_containers[fixed_texture_slot]

    if bg_container == nil then
        CHECK(false, "ShowBGService: SetBGTexture: fixed_texture_slot:",
        fixed_texture_slot , "is not created yet !!!")
        return
    end

    bg_container.bg_texture = texture
    bg_container.bg_sprite:GetMaterial():SetBaseColorTex(texture)
end

function ShowBGService:GetBGTexture(fixed_texture_slot)
    if fixed_texture_slot == nil then
        fixed_texture_slot = self.node_manager.INPUT_TEXTURE_0
    end

    local bg_container = self.bg_containers[fixed_texture_slot]

    if bg_container == nil then
        CHECK(false, "ShowBGService: SetBGTexture: fixed_texture_slot:",
                fixed_texture_slot , "is not created yet !!!")
        return nil
    end

    return bg_container.bg_texture
end

function ShowBGService:_OnAwake()

end

function ShowBGService:_OnStart()
    self:AddListener(hl.Hl_MSG_RUNTIME_RESET, self._OnRuntimeReset, self)

    for _, bg_container in pairs(self.bg_containers) do
        if bg_container.enable then
            bg_container.bg_sprite:SetVisible(true)
        end
    end
end

function ShowBGService:_OnStop()
    for _, bg_container in pairs(self.bg_containers) do
        bg_container.bg_sprite:SetVisible(false)
    end

    self:RemoveListener(hl.Hl_MSG_RUNTIME_RESET)
end

return ShowBGService