--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local GameFramework = import "GameFramework"
local NodeLayer = import "NodeLayer"
local Node = import "Node"

local StateRecordPlayNodeLayer = define ("StateRecordPlayNodeLayer", NodeLayer)

LOGD("StateRecordPlayNodeLayer has been defined, parent is ", NodeLayer)

function StateRecordPlayNodeLayer:_OnInitialize()
    self.tmp_reused_node_map = {}
    self.tmp_ordered_node_list = {}
end

function StateRecordPlayNodeLayer:OnRecordedEvent(msg)
    local message_id = msg:PopBegin()

    if message_id == hl.Hl_MSG_SCRIPT_SNAPSHOT then
        self:RestoreFromSnapshot(msg)
    elseif  message_id == hl.Hl_MSG_SCRIPT_NODE_SNAPSHOT then
        self:RestoreNodeFromSnapshot(msg)
    elseif message_id == hl.Hl_MSG_SCRIPT_NODE_LOGIC then
        local node_type = msg:PopString()
        local node_id = msg:PopInt32()
        local node, node_info = self:QueryNode(node_type, node_id)
        if node == nil or (node_info.node_layer ~= nil and node_info.node_layer ~= self) then
            LOGW("StateRecordPlayNodeLayer: OnRecordedEvent: Hl_MSG_SCRIPT_NODE_LOGIC:", node_type, node_id, "can not find or node layer is not self")
        else
            node:NodeEvent(msg, GameFramework.KMessage_Recorded)
        end
    end
end

function StateRecordPlayNodeLayer:RestoreFromSnapshot(snapshot_msg)
    self.tmp_reused_node_map = {}
    self.tmp_ordered_node_list = {}

    local node_num = snapshot_msg:PopInt32()

    local node_type = nil
    local node_id = nil

    local tmp_node = nil
    local tmp_node_info = nil
    local tmp_node_layer = nil

    local node_pos = 1
    for index = 1, node_num do
        node_type, node_id = Node:StaticReadName(snapshot_msg)

        tmp_node, tmp_node_info = self:QueryNode(node_type, node_id)

        if tmp_node == nil then
            tmp_node = self:_RequireNode(node_type, node_id)
            if tmp_node == nil then
                CHECK(false, "StateRecordPlayNodeLayer: RestoreFromSnapshot: _RequireNode:", node_type, "fail !!!")
            else
                tmp_node:SetMode(Node.kPassiveMode)
                self.node_manager:AddNodeImpl(tmp_node, node_pos, self)
                node_pos = node_pos + 1
                table.insert(self.tmp_ordered_node_list, tmp_node)
            end
        else
            tmp_node_layer = tmp_node_info.node_layer

            if tmp_node_layer == self then
                node_pos = node_pos + 1
                table.insert(self.tmp_ordered_node_list, tmp_node)
                tmp_node:SetNodeLayerPos(#self.tmp_ordered_node_list)
                -- for reused node, do not set its mode to passive any more ...
                self.tmp_reused_node_map[tmp_node] = true
            else
                LOGW("StateRecordPlayNodeLayer: RestoreFromSnapshot: Node:", node_type, node_id,
                        "is in other node layer:", tmp_node_layer:GetName(),
                        "can not add it in current layer:", self:GetName())
            end
        end
    end

    for _, old_node in ipairs(self.ordered_node_list) do
        if self.tmp_reused_node_map[old_node] == nil then
            LOGD("StateRecordPlayNodeLayer: RestoreFromSnapshot: Node:", old_node:GetType(),
                old_node:GetId(), "is not used in this frame, remove it ...")
            self.node_manager:RemoveNodeImpl(old_node:GetType(), old_node:GetId())
        end
    end

    self.ordered_node_list = self.tmp_ordered_node_list

    snapshot_msg:PopEnd()
end

function StateRecordPlayNodeLayer:_OnAddNode(node_type, node_id)
    CHECK(false, "StateRecordPlayNodeLayer: _OnAddNode:", node_type, node_id, "is not allowed !!!")
end

function StateRecordPlayNodeLayer:_OnRemoveNode(node_type, node_id)
    local node_pos = self.node_manager:RemoveNodeImpl(node_type, node_id)

    if nil == node_pos then
        return false
    end

    CHECK(node_pos > 0, "StateRecordPlayNodeLayer: _OnRemoveNode, Node's", node_type, node_id, "invalid pos:", node_pos, "!!!")

    LOGD("StateRecordPlayNodeLayer: _OnRemoveNode, Node's", node_type, node_id, "pos:", node_pos)
    table.remove(self.ordered_node_list, node_pos)

    self:UpdateNodesPos()

    return true
end

function StateRecordPlayNodeLayer:RestoreNodeFromSnapshot(snapshot_msg)
    local node_type, node_id = Node:StaticReadName(snapshot_msg)

    local node, node_info = self:QueryNode(node_type, node_id)

    if node == nil then
        LOGW("StateRecordPlayNodeLayer: RestoreNodeFromSnapshot: Node:", node_type, node_id, "does not exist !!! ")
        return
    end

    if node_info.node_layer ~= self then
        LOGW("StateRecordPlayNodeLayer: RestoreNodeFromSnapshot: Node:", node_type, node_id, "is in other node layer:", node_info.node_layer:GetName(), "can not restore it in current layer:", self:GetName())
        return
    end

    node:RestoreFromSnapshot(snapshot_msg)

    snapshot_msg:PopEnd()
end

return StateRecordPlayNodeLayer