--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local View = import "View"

local LinearLayout = define("LinearLayout", View)

function LinearLayout:ctor(node, name)
    View.ctor(self, node, name)

    self.is_horizontal = true

    self.padding_left = 0
    self.padding_right = 0
    self.padding_top = 0
    self.padding_bottom = 0
end

function LinearLayout:SetVertical(vertical)
    self.is_horizontal = not vertical
    self:Invalidate()
end

function LinearLayout:__PostLayout(dt)
    self.super.__PostLayout(self, dt)

    local final_width = 0
    local final_height = 0

    if self.is_horizontal then
        local cur_view_center_x = 0
        local prev_view = nil
        --TODO: consider height adjust
        --local max_y = -100000
        --local min_y = 100000
        --local cur_top = 0
        --local cur_bottom = 0
        for _, child_view in ipairs(self.child_views) do
            if prev_view == nil then
                cur_view_center_x = self.padding_left + child_view.margin_left + child_view.half_width
            else
                cur_view_center_x = cur_view_center_x + prev_view.half_width + prev_view.margin_right + child_view.margin_left + child_view.half_width
            end

            prev_view = child_view

            --local cur_height = child_view.height + child_view.margin_top + child_view.margin_bottom + self.padding_top + self.padding_bottom
            --
            --cur_top = child_view.center_y + cur_height / 2
            --cur_bottom = child_view.center_y - cur_height / 2
            --
            --if max_y < cur_top then
            --    max_y = cur_top
            --end
            --
            --if min_y > cur_bottom then
            --    min_y = cur_bottom
            --end

            child_view:SetCenterX(cur_view_center_x)
        end

        final_height = self.height

        if prev_view ~= nil then
            final_width = cur_view_center_x + prev_view.half_width + prev_view.margin_right + self.padding_right
        end

        for _, child_view in ipairs(self.child_views) do
            child_view:SetCenterX(child_view.center_x - final_width/2)
            if child_view:GetGravity() == View.kGravityCenter then
                child_view:SetCenterY(self.center_y)
            end
        end
    else
        --TODO: consider width adjust
        local cur_view_center_y = 0
        local prev_view = nil
        for _, child_view in ipairs(self.child_views) do
            if prev_view == nil then
                cur_view_center_y = -(self.padding_top + child_view.margin_top + child_view.half_height)
            else
                cur_view_center_y = cur_view_center_y - (prev_view.half_height + prev_view.margin_bottom + child_view.margin_top + child_view.half_height)
            end

            prev_view = child_view

            child_view:SetCenterY(cur_view_center_y)
        end

        final_width = self.width

        if prev_view ~= nil then
            final_height = -cur_view_center_y + prev_view.half_height + prev_view.margin_bottom + self.padding_bottom
        end

        for _, child_view in ipairs(self.child_views) do
            child_view:SetCenterY(child_view.center_y + final_height/2)
            if child_view:GetGravity() == View.kGravityCenter then
                child_view:SetCenterX(self.center_y)
            end
        end
    end

    self.width = final_width
    self.height = final_height
    self.half_width = self.width/2
    self.half_height = self.height/2

    LOGD("LinearLayout:", self.name ,"__PostLayout: final width/height:", final_width, final_height)
end

return LinearLayout