--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local View = import "View"

local TextView = define ("TextView", View)

local kFontSize = 18

TextView.kAlignLeft = 0
TextView.kAlignCenter = 1
TextView.kAlignRight = 2
TextView.kAlignTop = 0
TextView.kAlignBottom = 1

function TextView:ctor(node, name, window)
    View.ctor(self, node, name, window)

    self.texer_actor = self.attach_node:AllocActor(hl.Named(self.name .. "_texer_actor"))
    self.attach_node:AddToWorld(self.texer_actor)
    self.texer_actor:GetTransform():SetParent(self.actor:GetTransform())

    self.sys_text = self.attach_node:GetSystemText()

    self.max_width = 1000

    self.auto_resize = false

    self.font_size = kFontSize
    self.font_name = "normal"
    self.bold = 0
    self.incline = 0
    self.text_wrap = true
    self.horizontal_align = TextView.kAlignCenter
    self.vertical_align = TextView.kAlignCenter
    self.font_color = 0xFFFFFFFF
    self.border_width = 0
    self.border_color = 0

    self.text_texture = nil

    self.text = ""

    self.text_sprite = hl.Sprite_New()
    local material = hl.Sprite_CreateFlipMaterial()
    local render_state = material:GetRenderState()
    render_state.blend_src = hl.GPBlendMode_kSrcAlpha
    render_state.blend_dst = hl.GPBlendMode_kSrcAlphaInv
    self.text_sprite:SetMaterial(material)
    self.text_sprite:SetLayer(self.window.renderable_layer)
    self.texer_actor:AddComponent(self.text_sprite)
    self.attach_node:SetRenderLayer(self.text_sprite, self.attach_node.LAYER_OVERLAY_2D)
    table.insert(self.renderables, self.text_sprite)
end

function TextView:AutoResize(resize)
    self.auto_resize = resize
    self:Invalidate()
end

function TextView:__PostLayout(dt)
    self.super.__PostLayout(self, dt)

    self.sys_text.name = self.font_name
    self.sys_text.size = self.font_size
    self.sys_text.bold = self.bold
    self.sys_text.incline = self.incline
    self.sys_text.border = self.border_width
    self.sys_text.color = self.font_color
    self.sys_text.border_color = self.border_color
    self.sys_text.width = self.max_width - self.padding_left - self.padding_right
    self.sys_text.height = 0
    self.sys_text.horizontal_align = hl.SysText_kCenter
    self.sys_text.verical_align = self.vertical_align
    self.sys_text.word_wrap = self.text_wrap

    self.text_texture = self.sys_text:UpdateText(self.text)

    if self.auto_resize then
        if self.text_texture then
            -- TODO: call SetWidth()/SetHeight()
            self.width = self.text_texture:GetWidth() + self.padding_left + self.padding_right
            self.height = self.text_texture:GetHeight() + self.padding_top + self.padding_bottom
            self.half_width = self.width / 2
            self.half_height = self.height / 2
        else
            self.width = 0
            self.height = 0
            self.half_width = 0
            self.half_height = 0
        end
    else
        --TODO: vertical align
        if nil ~= self.text_texture then

            local cur_text_pos =  self.texer_actor:GetTransform():GetLocalPosition()

            if self.horizontal_align == TextView.kAlignLeft then
                self.texer_actor:GetTransform():SetLocalPosition(hl.Vec3(self.text_texture:GetWidth()/2 - self.width/2 + self.padding_left, cur_text_pos.y, cur_text_pos.z))
            elseif self.horizontal_align == TextView.kAlignRight then
                self.texer_actor:GetTransform():SetLocalPosition(hl.Vec3(self.width/2 - self.text_texture:GetWidth()/2 - self.padding_right, cur_text_pos.y, cur_text_pos.z))
            elseif self.horizontal_align == TextView.kAlignCenter then
                self.texer_actor:GetTransform():SetLocalPosition(hl.Vec3(0, cur_text_pos.y, cur_text_pos.z))
            end
        end
    end
end

function TextView:__PostDraw(dt)
    self.super.__PostDraw(self, dt)

    if self.text_texture then
        if self.text_texture:GetWidth() == 0 or self.text_texture:GetHeight() == 0 then
            self.text_sprite:SetVisible(false)
        else
            self.text_sprite:GetMaterial():SetBaseColorTex(self.text_texture)
            self.text_sprite:SetWidth(self.text_texture:GetWidth())
            self.text_sprite:SetHeight(self.text_texture:GetHeight())
        end
    else
        self.text_sprite:SetVisible(false)
    end
end

function TextView:SetMaxWidth(max_width)
    self.max_width = max_width
end

function TextView:SetFontBorderWidth(width)
    self.border_width = width
    self:Invalidate()
end

function TextView:SetFontBorderColor(color)
    self.border_color = color
    self:Invalidate()
end

function TextView:SetHorizontalAlign(align)
    self.horizontal_align = align
    self:Invalidate()
end

function TextView:SetVerticalAlign(align)
    self.vertical_align = align
    self:Invalidate()
end

function TextView:SetFontName(name)
    self.font_name = name
    self:Invalidate()
end

function TextView:EnableWrap(wrap)
    self.text_wrap = wrap
    self:Invalidate()
end

function TextView:SetBold(bold)
    self.bold = bold
    self:Invalidate()
end

function TextView:SetFontSize(font_size)
    self.font_size = font_size
    self:Invalidate()
end

function TextView:OrderSelf(begin_order)
    local new_begin_order = View.OrderSelf(self, begin_order)

    self.text_sprite:SetOrder(new_begin_order)

    return new_begin_order + 1
end

function TextView:SetText(text)
    self.text = text
    self:Invalidate()
end

function TextView:SetFontColor(color)
    self.font_color = color
    self:Invalidate();
end

return TextView