--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
require "UI/View"
require "UI/TextView"
require "UI/LinearLayout"

local View = import "View"

local Window = define "Window"

Window.kDefaultRenderableLayer = 0

function Window:ctor(node, name, renderable_layer)
    self.attach_node = node
    self.name = name
    if nil == renderable_layer then
        self.renderable_layer = Window.kDefaultRenderableLayer
    else
        self.renderable_layer = renderable_layer
    end

    self.begin_order = 0
    self.end_order = 0

    self.root_view = View:New(node, "root_view", self)

    --FIXME: we may need a alpha texture ....
    self.root_view.background_sprite:SetVisible(false)

    local runtime_width = hl.RuntimeWidth(hl.kSCREEN_EQUAL)
    local runtime_height = hl.RuntimeHeight(hl.kSCREEN_EQUAL)

    self.root_view:SetWidth(runtime_width)
    self.root_view:SetHeight(runtime_height)
    self.root_view:SetTouchable(true)

    self.listeners = {}

    self:_AddListener(hl.Hl_MSG_RUNTIME_RESET, self._OnRuntimeReset, self)
end

function Window:_OnRuntimeReset()
    local runtime_width = hl.RuntimeWidth(hl.kSCREEN_EQUAL)
    local runtime_height = hl.RuntimeHeight(hl.kSCREEN_EQUAL)

    self.root_view:SetWidth(runtime_width)
    self.root_view:SetHeight(runtime_height)
end

function Window:_AddListener(event, listener, holder)
    if self.listeners[event] ~= nil then
        CHECK(false, "Window ", self.name, "event", event, "listener already added !!!")
        return
    end

    LOGD("Window ", self.name, "event", event, "listener added ...")
    self.listeners[event] = listener

    self.attach_node.node_context.game_framework:AddListener(event, listener, holder)
end

function Window:_RemoveListener(event)
    if self.listeners[event] == nil then
        CHECK(false, "Window ", self.name, "event", event, "listener not added yet !!!")
        return
    end

    LOGD("Window ", self.name, "event", event, "listener removed ...")

    self.attach_node.node_context.game_framework:RemoveListener(event, self.listeners[event])
    self.listeners[event] = nil
end

function Window:_ReleaseListeners()
    for event, listener in pairs(self.listeners) do
        if listener ~= nil then
            LOGD("Window ", self.name," remove event", event, "listener")
            self.attach_node.node_context.game_framework:RemoveListener(event, listener)
        end
    end

    self.listeners = nil
end

function Window:GetRootView()
    return self.root_view
end

function Window:SetBeginOrder(begin_order)
    self.begin_order = begin_order
    self:OrderView()
end

function Window:OrderView()
    self.end_order = self.root_view:Order(self.begin_order)
    LOGD("Window", self.name, "order view: end order", self.end_order)
end

function Window:Update(dt)
    self.root_view:Layout(dt)
    self.root_view:Draw(dt)
end

function Window:OnMouseEvent(action, x, y)
    self.root_view:OnMouseEvent(action, x, y)
end

function Window:OnTouchEvent(action, touches)
    self.root_view:OnTouchEvent(action, touches)
end

function Window:Destroy()
    self:_ReleaseListeners()
end

return Window